import copy
import logging as log
import sbpy.activity as sba
from sbpy.data import Phys
from dataclasses import dataclass
from pyvectorial_au.pre_model_processing.timedependentproduction import (
    make_time_dependence_function,
)

from pyvectorial_au.model_input.vectorial_model_config import (
    BinnedProductionTimeVariation,
    VectorialModelConfig,
)
from pyvectorial_au.model_output.vectorial_model_result import (
    VectorialModelResult,
    FragmentSputterSpherical,
)


"""
    For interfacing with the python version of the vectorial model
"""


@dataclass
class PythonModelExtraConfig:
    print_progress: bool = False


def parent_molecule_phys_from_vmc(vmc: VectorialModelConfig) -> Phys:
    p = vmc.parent
    p_dict = {
        "tau_d": p.tau_d,
        "tau_T": p.tau_T,
        "v_outflow": p.v_outflow,
        "sigma": p.sigma,
    }

    return Phys.from_dict(p_dict)


def fragment_molecule_phys_from_vmc(vmc: VectorialModelConfig) -> Phys:
    f = vmc.fragment
    f_dict = {"tau_T": f.tau_T, "v_photo": f.v_photo}

    return Phys.from_dict(f_dict)


def run_python_vectorial_model(
    vmc: VectorialModelConfig, extra_config: PythonModelExtraConfig
) -> VectorialModelResult:
    """
    Takes a VectorialModelConfig and builds the time dependence function
    specified, then runs the sbpy vectorial model to return a coma object
    """

    # build parent and fragment inputs
    parent = parent_molecule_phys_from_vmc(vmc=vmc)
    fragment = fragment_molecule_phys_from_vmc(vmc=vmc)

    coma = None
    q_t = None

    # handle binned production as a special case
    if isinstance(vmc.production.time_variation, BinnedProductionTimeVariation):
        # call the binned production constructor that mimics the fortran version
        coma = sba.VectorialModel.binned_production(
            qs=vmc.production.time_variation.q,  # type: ignore
            parent=parent,
            fragment=fragment,
            ts=vmc.production.time_variation.times_at_productions,  # type: ignore
            radial_points=vmc.grid.radial_points,
            angular_points=vmc.grid.angular_points,
            radial_substeps=vmc.grid.radial_substeps,
            print_progress=extra_config.print_progress,
        )
    else:
        if vmc.production.time_variation:
            q_t = make_time_dependence_function(vmc=vmc)
        else:
            q_t = None
            log.info(
                "No valid time dependence specified, assuming steady production of %s",
                vmc.production.base_q,
            )

        coma = sba.VectorialModel(
            base_q=vmc.production.base_q,
            q_t=q_t,
            parent=parent,
            fragment=fragment,
            radial_points=vmc.grid.radial_points,
            angular_points=vmc.grid.angular_points,
            radial_substeps=vmc.grid.radial_substeps,
            print_progress=extra_config.print_progress,
        )

    return VectorialModelResult(
        volume_density_grid=coma.vmr.volume_density_grid,
        volume_density=coma.vmr.volume_density,
        fragment_sputter=FragmentSputterSpherical(
            rs=coma.vmr.fragment_sputter.rs,
            thetas=coma.vmr.fragment_sputter.thetas,
            fragment_density=coma.vmr.fragment_sputter.fragment_density,
        ),
        collision_sphere_radius=coma.vmr.collision_sphere_radius,
        max_grid_radius=coma.vmr.max_grid_radius,
        coma_radius=coma.vmr.coma_radius,
        column_density_grid=coma.vmr.column_density_grid,
        column_density=coma.vmr.column_density,
        volume_density_interpolation=coma.vmr.volume_density_interpolation,
        column_density_interpolation=coma.vmr.column_density_interpolation,
        coma=coma,
    )


def vmr_from_sbpy_coma(coma_orig: sba.VectorialModel) -> VectorialModelResult:
    """Takes a coma object from the sbpy python version and extracts a VectorialModelResult"""
    coma = copy.deepcopy(coma_orig)

    vdg = coma.vmr.volume_density_grid
    vd = coma.vmr.volume_density
    cdg = coma.vmr.column_density_grid
    cd = coma.vmr.column_density
    vdi = coma.vmr.volume_density_interpolation
    cdi = coma.vmr.column_density_interpolation

    fs = FragmentSputterSpherical(
        rs=coma.vmr.fragment_sputter.rs,
        thetas=coma.vmr.fragment_sputter.thetas,
        fragment_density=coma.vmr.fragment_sputter.fragment_density,
    )

    csr = coma.vmr.collision_sphere_radius
    mgr = coma.vmr.max_grid_radius
    cr = coma.vmr.coma_radius

    return VectorialModelResult(
        volume_density_grid=vdg,
        volume_density=vd,
        fragment_sputter=fs,
        collision_sphere_radius=csr,
        max_grid_radius=mgr,
        coma_radius=cr,
        column_density_grid=cdg,
        column_density=cd,
        volume_density_interpolation=vdi,
        column_density_interpolation=cdi,
    )



import pathlib
from sqlalchemy import NullPool, create_engine
from pyvectorial_au.db.vectorial_model_cache import VMCachedBase


def initialize_vectorial_model_cache(
    vectorial_model_cache_dir: pathlib.Path,
    vectorial_model_cache_filename: pathlib.Path = pathlib.Path("vmcache.sqlite3"),
) -> pathlib.Path:
    """
    On startup we call this before any work is done to make sure the db has the proper
    tables and columns, and return the full file path for use later
    """

    full_db_path = vectorial_model_cache_dir.joinpath(vectorial_model_cache_filename)
    vm_cache_engine = create_engine(f"sqlite:///{full_db_path}", poolclass=NullPool)

    VMCachedBase.metadata.create_all(bind=vm_cache_engine)

    return full_db_path



import logging as log
from enum import Enum, auto

import numpy as np
import astropy.units as u
from astropy.units.quantity import Quantity

from pyvectorial_au.model_input.vectorial_model_config import (
    FragmentMolecule,
    ParentMolecule,
    VectorialModelConfig,
)


"""
    There are different approaches to scaling parameters with heliocentric distance.
    These functions provide some of the common ones dealt with during testing of the vectorial model
"""


class VmcTransform(Enum):
    cochran_schleicher_93 = auto()
    fortran_festou = auto()


def apply_input_transform(
    vmc: VectorialModelConfig, r_h: Quantity, xfrm: VmcTransform
) -> VectorialModelConfig:
    if xfrm == VmcTransform.cochran_schleicher_93:
        return _apply_cochran_schleicher_93(vmc, r_h)
    elif xfrm == VmcTransform.fortran_festou:
        return _apply_festou_fortran(vmc, r_h)


def _apply_cochran_schleicher_93(
    vmc_orig: VectorialModelConfig,
    r_h: Quantity,
) -> VectorialModelConfig:
    rh = r_h.to_value(u.AU)
    rhsquared = rh**2  # type: ignore
    sqrh = np.sqrt(rh)

    new_parent = ParentMolecule(
        v_outflow_kms=0.85 / sqrh,  # type: ignore
        tau_d_s=vmc_orig.parent.tau_d_s * rhsquared,
        tau_T_s=vmc_orig.parent.tau_T_s * rhsquared,
        sigma_cm_sq=vmc_orig.parent.sigma_cm_sq,
    )
    new_fragment = FragmentMolecule(
        tau_T_s=vmc_orig.fragment.tau_T_s * rhsquared,
        v_photo_kms=vmc_orig.fragment.v_photo_kms,
    )

    vmc = vmc_orig.model_copy(
        update={"parent": new_parent, "fragment": new_fragment}, deep=True
    )

    log.debug("Applying transform cochran_schleicher_93...")
    log.debug("Effect of transform at %s AU:", rh)
    log.debug(
        "Parent outflow: %s --> %s", vmc_orig.parent.v_outflow, vmc.parent.v_outflow
    )
    log.debug("Parent tau_d: %s --> %s", vmc_orig.parent.tau_d, vmc.parent.tau_d)
    log.debug("Parent tau_T: %s --> %s", vmc_orig.parent.tau_T, vmc.parent.tau_T)
    log.debug("Fragment tau_T: %s --> %s", vmc_orig.fragment.tau_T, vmc.fragment.tau_T)

    return vmc


def _apply_festou_fortran(
    vmc_orig: VectorialModelConfig, r_h: Quantity
) -> VectorialModelConfig:
    rh = r_h.to(u.AU).value
    rhsquared = rh**2

    new_parent = vmc_orig.parent.model_copy(
        update={
            "tau_d_s": vmc_orig.parent.tau_d_s * rhsquared,
            "tau_T_s": vmc_orig.parent.tau_T_s * rhsquared,
        },
        deep=True,
    )
    new_fragment = vmc_orig.fragment.model_copy(
        update={"tau_T_s": vmc_orig.fragment.tau_T_s * rhsquared}
    )

    vmc = vmc_orig.model_copy(
        update={"parent": new_parent, "fragment": new_fragment}, deep=True
    )

    log.debug("Applying transform fortran_festou...")
    log.debug("Effect of transform at %s AU:", rh)
    log.debug("\tParent tau_d: %s --> %s", vmc_orig.parent.tau_d, vmc.parent.tau_d)
    log.debug("\tParent tau_T: %s --> %s", vmc_orig.parent.tau_T, vmc.parent.tau_T)
    log.debug(
        "\tFragment tau_T: %s --> %s", vmc_orig.fragment.tau_T, vmc.fragment.tau_T
    )

    return vmc



from typing import Union
import numpy as np

from pyvectorial_au.model_output.fragment_sputter import (
    FragmentSputterSpherical,
    FragmentSputterCartesian,
    FragmentSputterPolar,
)


def fragment_sputter_to_cartesian(
    fsp: Union[FragmentSputterSpherical, FragmentSputterPolar]
) -> FragmentSputterCartesian:
    """
    Fragment sputter information is generated by the model in terms of (r, theta, density), so this converts to (x, y, density)
    where theta is a spherical polar angle
    """
    return FragmentSputterCartesian(
        xs=fsp.rs * np.sin(fsp.thetas),
        ys=fsp.rs * np.cos(fsp.thetas),
        fragment_density=fsp.fragment_density,
    )


def fragment_sputter_to_polar(fsp: FragmentSputterSpherical) -> FragmentSputterPolar:
    """Utility function to change the typing of given fragment sputter"""
    return FragmentSputterPolar(
        rs=fsp.rs, thetas=fsp.thetas, fragment_density=fsp.fragment_density
    )


def mirror_fragment_sputter(
    fsp: Union[FragmentSputterPolar, FragmentSputterCartesian]
) -> Union[FragmentSputterPolar, FragmentSputterCartesian]:
    """
    The sputter in (r, theta, density) in spherical coordinate format occupies the positive x-axis only due to the azimuthal symmetry of the problem.
    When we convert that to other coordinate systems, we might want to include the negative x-axis by mirroring the fragment sputter around x = 0,
    which this function provides.
    The return type is the same type as the input
    """
    if isinstance(fsp, FragmentSputterPolar):
        fsp.rs = np.append(fsp.rs, fsp.rs)
        fsp.thetas = np.append(fsp.thetas, -1 * fsp.thetas)
        fsp.fragment_density = np.append(fsp.fragment_density, fsp.fragment_density)
    elif isinstance(fsp, FragmentSputterCartesian):
        fsp.xs = np.append(fsp.xs, -1 * fsp.xs)
        fsp.ys = np.append(fsp.ys, fsp.ys)
        fsp.fragment_density = np.append(fsp.fragment_density, fsp.fragment_density)

    return fsp



import copy

import numpy as np
import astropy.units as u
import scipy.interpolate

import matplotlib.pyplot as plt
import matplotlib.cm as cmx
from matplotlib.colors import Normalize

from pyvectorial_au.model_output.vectorial_model_result import VectorialModelResult
from pyvectorial_au.model_output.fragment_sputter import (
    FragmentSputterSpherical,
    FragmentSputterPolar,
)
from pyvectorial_au.post_model_processing.fragment_sputter_transform import (
    mirror_fragment_sputter,
    fragment_sputter_to_cartesian,
)


"""
    Functions for using matplotlib to plot various data contained in VectorialModelResults
"""

# color palette
myred = "#c74a77"
mybred = "#dbafad"
mygreen = "#afac7c"
mybgreen = "#dbd89c"
mypeach = "#dbb89c"
mybpeach = "#e9d4c3"
myblue = "#688894"
mybblue = "#a4b7be"
myblack = "#301e2a"
mybblack = "#82787f"
mywhite = "#d8d7dc"
mybwhite = "#e7e7ea"


def _find_cdens_inflection_points(vmr: VectorialModelResult) -> np.ndarray:
    """
    Look for changes in sign of second derivative of the column density,
    given a VectorialModelResult and return a list of inflection points
    """

    # choose a sample of radii to test for changes in sign of second derivative
    xs = np.linspace(0, 5e8, num=100)
    concavity = vmr.column_density_interpolation.derivative(nu=2)  # type: ignore
    # compute values of the second derivative at our sample points
    ys = concavity(xs)

    # Array of 1s or 0s marking if the sign changed from one element to the next
    sign_changes = (np.diff(np.sign(ys)) != 0) * 1

    # Manually remove the weirdness near the nucleus and rezize the array
    sign_changes[0] = 0
    sign_changes = np.resize(sign_changes, 100)

    inflection_points = xs * sign_changes
    # Only want non-zero elements
    inflection_points = inflection_points[inflection_points > 0]

    # Only want inflection points outside the collision sphere
    csphere_radius = vmr.collision_sphere_radius.to_value(u.m)
    inflection_points = inflection_points[inflection_points > csphere_radius]  # type: ignore

    # our model's interpolator function deals in meters, so tag results with the proper units
    inflection_points = inflection_points * u.m  # type: ignore
    return inflection_points


def mpl_mark_inflection_points(vmr: VectorialModelResult, ax, **kwargs) -> None:
    # Find possible inflection points in the column density
    for ipoint in _find_cdens_inflection_points(vmr):
        ax.axvline(x=ipoint, **kwargs)


def mpl_mark_collision_sphere(vmr: VectorialModelResult, ax, **kwargs) -> None:
    # Mark the beginning of the collision sphere
    ax.axvline(x=vmr.collision_sphere_radius, **kwargs)


# ax is an axis from a matplotlib figure
def mpl_volume_density_plot(
    vmr: VectorialModelResult, ax, dist_units=u.m, vdens_units=1 / u.m**3, **kwargs  # type: ignore
) -> None:
    xs = vmr.volume_density_grid.to(dist_units)  # type: ignore
    ys = vmr.volume_density.to(vdens_units)  # type: ignore

    ax.scatter(xs, ys, **kwargs)


def mpl_volume_density_interpolation_plot(
    vmr: VectorialModelResult, ax, dist_units=u.m, vdens_units=1 / u.m**3, **kwargs  # type: ignore
) -> None:
    if vmr.volume_density_interpolation is None:
        return

    # model's interpolation function needs meters in, gives output in 1/m**3
    ys = (
        vmr.volume_density_interpolation(vmr.volume_density_grid.to_value(u.m)) / u.m**3  # type: ignore
    )
    ax.plot(vmr.volume_density_grid.to(dist_units), ys.to(vdens_units), **kwargs)  # type: ignore


def mpl_column_density_plot(
    vmr: VectorialModelResult, ax, dist_units=u.m, cdens_units=1 / u.m**2, **kwargs  # type: ignore
) -> None:
    xs = vmr.column_density_grid.to(dist_units)  # type: ignore
    ys = vmr.column_density.to(cdens_units)  # type: ignore

    ax.scatter(xs, ys, **kwargs)


def mpl_column_density_interpolation_plot(
    vmr: VectorialModelResult, ax, dist_units=u.m, cdens_units=1 / u.m**2, **kwargs  # type: ignore
) -> None:
    if vmr.column_density_interpolation is None:
        return
    # model's interpolation function needs meters in, gives output in 1/m**2
    ys = (
        vmr.column_density_interpolation(vmr.column_density_grid.to_value(u.m)) / u.m**2  # type: ignore
    )
    ax.plot(vmr.column_density_grid.to(dist_units), ys.to(cdens_units), **kwargs)  # type: ignore


def mpl_column_density_plot_3d(
    vmr: VectorialModelResult,
    ax,
    center=(0, 0) * u.m,  # type: ignore
    width=200000 * u.km,  # type: ignore
    height=200000 * u.km,  # type: ignore
    divisions=100,
    dist_units=u.m,
    cdens_units=1 / u.m**2,  # type: ignore
    **kwargs
) -> None:
    xmin_m, ymin_m = np.subtract(
        center.to_value(u.m), (width.to_value(u.m) / 2, height.to_value(u.m) / 2)
    )
    xmax_m, ymax_m = np.add(
        (xmin_m, ymin_m), (width.to_value(u.m), height.to_value(u.m))
    )
    xs_m = np.linspace(xmin_m, xmax_m, num=divisions)
    ys_m = np.linspace(ymin_m, ymax_m, num=divisions)

    xmesh_m, ymesh_m = np.meshgrid(xs_m, ys_m)
    zmesh = vmr.column_density_interpolation(np.sqrt(xmesh_m**2 + ymesh_m**2)) / u.m**2  # type: ignore
    xmesh = xmesh_m * u.m
    ymesh = ymesh_m * u.m

    vmin = np.min(zmesh).to_value(cdens_units)
    vmax = np.max(zmesh).to_value(cdens_units)

    ax.plot_surface(
        xmesh.to(dist_units),
        ymesh.to(dist_units),
        zmesh.to(cdens_units),
        vmin=vmin,
        vmax=vmax,
        cmap=cmx.get_cmap("viridis"),
        norm="log",
        **kwargs,
    )
    ax.set_zlim(bottom=0)


def mpl_fragment_sputter_contour_plot(
    vmr,
    ax,
    dist_units=u.km,
    sputter_units=1 / u.cm**3,
    within_r=1000 * u.km,  # type: ignore
    min_r=0 * u.km,  # type: ignore
    max_angle=np.pi,
    mirrored=False,
    show_outflow_axis=True,
    **kwargs
) -> None:
    fragment_sputter = copy.deepcopy(vmr.fragment_sputter)

    within_max_angle = fragment_sputter.thetas < max_angle
    fragment_sputter.rs = fragment_sputter.rs[within_max_angle]
    fragment_sputter.thetas = fragment_sputter.thetas[within_max_angle]
    fragment_sputter.fragment_density = fragment_sputter.fragment_density[
        within_max_angle
    ]

    if mirrored:
        fragment_sputter = mirror_fragment_sputter(fragment_sputter)

    if isinstance(fragment_sputter, FragmentSputterPolar) or isinstance(
        fragment_sputter, FragmentSputterSpherical
    ):
        fragment_sputter = fragment_sputter_to_cartesian(fragment_sputter)

    xs = fragment_sputter.xs.to(dist_units)  # type: ignore
    ys = fragment_sputter.ys.to(dist_units)  # type: ignore
    zs = fragment_sputter.fragment_density.to(sputter_units)  # type: ignore

    # within_limit = np.sqrt(xs**2 + ys**2) < within_r
    within_limit = np.logical_and(
        np.sqrt(xs**2 + ys**2) < within_r, np.sqrt(xs**2 + ys**2) > min_r
    )
    xs = xs[within_limit]
    ys = ys[within_limit]
    zs = zs[within_limit]

    if show_outflow_axis:
        # highlight the outflow axis, along positive y
        origin = [0, 0, 0] * dist_units  # type: ignore
        outflow_max = [0, np.max(ys.to_value(dist_units)), 0] * dist_units  # type: ignore
        ax.plot(origin, outflow_max, color=myblue, lw=2, label="outflow axis")

    x_mesh, y_mesh = np.meshgrid(np.unique(xs), np.unique(ys))
    print("Interpolating surface...")
    fs_rbf = scipy.interpolate.Rbf(xs, ys, zs, function="cubic")
    print("Computing interploted values...")
    frag_mesh = fs_rbf(x_mesh, y_mesh)
    print("Graphing..")

    ax.contourf(
        x_mesh,
        y_mesh,
        frag_mesh,
        levels=np.arange(np.min(frag_mesh), np.max(frag_mesh), 50),
        cmap=cmx.get_cmap("viridis"),
        norm="log",
        **kwargs,
    )
    ax.contour(
        x_mesh,
        y_mesh,
        frag_mesh,
        levels=np.arange(np.min(frag_mesh), np.max(frag_mesh), 50),
        colors="black",
        linewidths=0.5,
        alpha=0.5,
    )
    ax.set_aspect("equal")


def mpl_fragment_sputter_plot(
    vmr: VectorialModelResult,
    ax,
    dist_units=u.m,
    sputter_units=1 / u.m**3,  # type: ignore
    within_r=1000 * u.km,  # type: ignore
    mirrored: bool = False,
    show_outflow_axis: bool = True,
    **kwargs
) -> None:
    fragment_sputter = vmr.fragment_sputter

    if isinstance(fragment_sputter, FragmentSputterPolar) or isinstance(
        fragment_sputter, FragmentSputterSpherical
    ):
        fragment_sputter = fragment_sputter_to_cartesian(fragment_sputter)

    if mirrored:
        fragment_sputter = mirror_fragment_sputter(fragment_sputter)

    xs = fragment_sputter.xs.to(dist_units)  # type: ignore
    ys = fragment_sputter.ys.to(dist_units)  # type: ignore
    zs = fragment_sputter.fragment_density.to(sputter_units)  # type: ignore

    within_limit = np.sqrt(xs**2 + ys**2) < within_r

    xs = xs[within_limit]
    ys = ys[within_limit]
    zs = zs[within_limit]

    colors_map = "viridis"
    cm = plt.get_cmap(colors_map)
    cNorm = Normalize(vmin=np.min(zs.value), vmax=np.max(zs.value))
    scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=cm)

    # highlight the outflow axis, along positive y
    if show_outflow_axis:
        origin = [0, 0, 0] * dist_units  # type: ignore
        outflow_max = [0, np.max(ys.to_value(dist_units)), 0] * dist_units  # type: ignore
        ax.plot(origin, outflow_max, color=myblue, lw=2, label="outflow axis")

    ax.scatter(xs, ys, zs, c=kwargs.get("color", scalarMap.to_rgba(zs.value)))



import numpy as np
import matplotlib.pyplot as plt
import sbpy.activity as sba

from pyvectorial_au.haser.haser_params import HaserParams

"""
    Utility functions for Haser-related plotting
"""


def plot_haser_column_density(hps: HaserParams, ax: plt.Axes, rs: np.ndarray) -> None:  # type: ignore
    """Take HaserParams and plot the column density along the values in the array rs"""

    coma = sba.Haser(Q=hps.q, v=hps.v_outflow, parent=hps.gamma_p, daughter=hps.gamma_d)
    ax.plot(rs, coma.column_density(rs))



from dataclasses import dataclass
from typing import List, Callable
from functools import partial

import numpy as np
import astropy.units as u
from scipy import special, optimize

from pyvectorial_au.haser.haser_params import HaserParams


"""
Functions for taking column density data as a function or radius and fitting to a Haser model

Implemented: Finding best-fit Q for Haser model given HaserParams (scale lengths, outflow velocity)
"""


def _haser_column_density(rho_m, q_s, v_ms, p_m, f_m) -> Callable:
    """
    Calculates the Haser column density at rho with given parameters to model,
    no astropy units attached, all quantities in meters and seconds
    """
    sigma = q_s / (v_ms * rho_m * 2 * np.pi)
    sigma *= (f_m / (p_m - f_m)) * (
        special.iti0k0(rho_m / f_m)[1] - special.iti0k0(rho_m / p_m)[1]
    )
    return sigma


def _make_haser_column_density_q(hps: HaserParams) -> Callable:
    """
    Takes HaserParams and returns a column density function with Q as a fitting parameter

    Fitting function takes impact parameter in meters and returns column density in 1/m**2
    """
    p_m = hps.gamma_p.to_value("m")  # type: ignore
    f_m = hps.gamma_d.to_value("m")  # type: ignore
    v_ms = hps.v_outflow.to_value("m/s")  # type: ignore

    f = partial(_haser_column_density, v_ms=v_ms, p_m=p_m, f_m=f_m)
    return f


@dataclass
class HaserFitResult:
    """
    Dataclass for returning the results of a fit of Haser model column density with data
    """

    # Function that was fitted
    fitting_function: Callable
    # np.ndarray of best-fit parameters
    fitted_params: List
    # np.ndarry of parameter covariances
    covariances: List


def haser_params_from_full_fit_result(hfr: HaserFitResult) -> HaserParams:
    return HaserParams(
        q=hfr.fitted_params[0] / u.s,
        v_outflow=hfr.fitted_params[1] * (u.m / u.s),
        gamma_p=hfr.fitted_params[2] * u.m,
        gamma_d=hfr.fitted_params[3] * u.m,
    )


def haser_q_fit_from_column_density(
    q_guess: u.Quantity, hps: HaserParams, rs: np.ndarray, cds: np.ndarray
) -> HaserFitResult:
    # Take HaserParams in hps along with a guess for Q to run Haser models until a best fit is found
    # to the data in rs, cds

    # We're performing the fit to find Q, so warn if the user filled in a value for Q already
    if hps.q is not None:
        print("Warning: do_haser_q_fit received non-empty production in HaserParams")

    hcd = _make_haser_column_density_q(hps)
    popt, pcov = optimize.curve_fit(hcd, rs, cds, p0=[q_guess.to_value("1/s")])
    return HaserFitResult(fitting_function=hcd, fitted_params=popt, covariances=pcov)


def haser_full_fit_from_column_density(
    q_guess, v_guess, parent_guess, fragment_guess, rs, cds
) -> HaserFitResult:
    # given column density data in rs, cds, find the Haser parameter set that
    # best fits the data

    qg = q_guess.to_value("1/s")
    vg = v_guess.to_value("m/s")
    pg = parent_guess.to_value("m")
    fg = fragment_guess.to_value("m")

    hcd = _haser_column_density
    popt, pcov = optimize.curve_fit(hcd, rs, cds, p0=[qg, vg, pg, fg])
    return HaserFitResult(fitting_function=hcd, fitted_params=popt, covariances=pcov)


# @dataclass
# class HaserScaleLengthSearchResult:
#     """
#     Dataclass for compiling results of which parent, fragment scale length pair best match
#     a given vectorial model result, with 'match' meaning they agree on the production
#     """
#
#     # regular arrays
#     parent_gammas: np.array = None
#     fragment_gammas: np.array = None
#     fitted_qs: np.array = None
#
#     # meshgrids
#     p_mesh: np.array = None
#     f_mesh: np.array = None
#     q_mesh: np.array = None
#
#     # measure of agreement on total production
#     # 0 is best: 'distance' of fitted production away from vectorial production
#     agreements: np.array = None
#     a_mesh: np.array = None
#
#     # best fits for this search
#     best_params: HaserParams = None
#
#
# def find_best_haser_scale_lengths_q(
#     vmc, vmr, parent_gammas: np.array, fragment_gammas: np.array
# ) -> HaserScaleLengthSearchResult:
#     # Takes a finished vectorial model, fits Haser models of various scale lengths,
#     # and looks for the (parent, fragment) scale length pair that agrees with the vectorial model's
#     # input production
#
#     num_parent_gammas = np.size(parent_gammas)
#     num_fragment_gammas = np.size(fragment_gammas)
#
#     fitting_results = []
#     for parent_gamma in parent_gammas:
#         for fragment_gamma in fragment_gammas:
#             hps = HaserParams(
#                 q=None,
#                 v_outflow=vmc.parent.v_outflow,
#                 gamma_p=parent_gamma,
#                 gamma_d=fragment_gamma,
#             )
#             hsr = haser_q_fit_from_column_density(
#                 q_guess=vmc.production.base_q,
#                 hps=hps,
#                 rs=vmr.column_density_grid,
#                 cds=vmr.column_density,
#             )
#             # output is a list of the form [[parent, fragment, q_fitted], ...]
#             fitting_results.append(
#                 [
#                     parent_gamma.to_value("km"),
#                     fragment_gamma.to_value("km"),
#                     hsr.fitted_params[0],
#                 ]
#             )
#
#     fdata = np.array(fitting_results).reshape(
#         (num_parent_gammas * num_fragment_gammas, 3)
#     )
#
#     result = HaserScaleLengthSearchResult
#
#     # take columns of each variable
#     result.parent_gammas = fdata[:, 0]
#     result.fragment_gammas = fdata[:, 1]
#     result.fitted_qs = fdata[:, 2]
#
#     # generate meshgrids for plotting etc.
#     result.p_mesh, result.f_mesh = np.meshgrid(
#         np.unique(result.parent_gammas), np.unique(result.fragment_gammas)
#     )
#     q_rbf = scipy.interpolate.Rbf(
#         result.parent_gammas, result.fragment_gammas, result.fitted_qs, function="cubic"
#     )
#     result.q_mesh = q_rbf(result.p_mesh, result.f_mesh)
#
#     # find how much haser production agrees with given vectorial production
#     result.agreements = np.sqrt(
#         (result.fitted_qs / vmc.production.base_q.to_value("1/s") - 1) ** 2
#     )
#     a_rbf = scipy.interpolate.Rbf(
#         result.parent_gammas,
#         result.fragment_gammas,
#         result.agreements,
#         function="cubic",
#     )
#     result.a_mesh = a_rbf(result.p_mesh, result.f_mesh)
#
#     # find index of minimum difference (best fit of productions)
#     best = np.unravel_index(np.argmin(result.a_mesh, axis=None), result.a_mesh.shape)
#     result.best_params = HaserParams(
#         q=result.q_mesh[best],
#         v_outflow=vmc.parent.v_outflow,
#         gamma_p=result.p_mesh[best],
#         gamma_d=result.f_mesh[best],
#     )
#
#     return result



from functools import partial
from multiprocessing import Pool
import time
import pathlib
import importlib.metadata as impm
from typing import Union, Optional, List

from pyvectorial_au.db.cache_io import get_saved_model_from_cache, save_model_to_cache
from pyvectorial_au.model_input.vectorial_model_config import VectorialModelConfig
from pyvectorial_au.backends.python_version import (
    PythonModelExtraConfig,
    run_python_vectorial_model,
)
from pyvectorial_au.backends.fortran_version import (
    FortranModelExtraConfig,
    run_fortran_vectorial_model,
)
from pyvectorial_au.backends.rust_version import (
    RustModelExtraConfig,
    run_rust_vectorial_model,
)
from pyvectorial_au.encoding.encoding_and_hashing import (
    pickle_to_base64,
    vmc_to_sha256_digest,
)
from pyvectorial_au.model_output.vectorial_model_calculation import (
    EncodedVMCalculation,
    VMCalculation,
)


def verbose_print(x, verbose: bool) -> None:
    if verbose:
        print(x)


def run_vectorial_model_single(
    vmc: VectorialModelConfig,
    extra_config: Union[
        PythonModelExtraConfig, FortranModelExtraConfig, RustModelExtraConfig
    ] = PythonModelExtraConfig(print_progress=False),
    vectorial_model_cache_path: Optional[pathlib.Path] = None,
) -> EncodedVMCalculation:
    """
    Given a vectorial model config and database path, look up the result in the database and return
    the already-run model.  If it is not in the database, run it and add it to the db.
    If no db path is specified, just run the model and return the result.
    This was meant to run in its own process through multiprocessing so it must
    create its own database engine.
    """
    # attempt to load model if we find it in the database
    if vectorial_model_cache_path:
        evmc = get_saved_model_from_cache(
            vmc=vmc, vectorial_model_cache_path=vectorial_model_cache_path
        )
        if evmc:
            return evmc

    model_start_time = time.time()
    model_function = None

    # TODO: rust version is hard-coded here - look for a better way to do this - maybe use the bin path to ask the rust executable with 'rust_vectorial_model --version' - need to add this functionality to rust version
    if isinstance(extra_config, PythonModelExtraConfig):
        vectorial_model_backend = "python"
        vectorial_model_version = impm.version("sbpy")
        model_function = run_python_vectorial_model
    elif isinstance(extra_config, FortranModelExtraConfig):
        vectorial_model_backend = "fortran"
        vectorial_model_version = "1.0.0"
        model_function = run_fortran_vectorial_model
    elif isinstance(extra_config, RustModelExtraConfig):
        vectorial_model_backend = "rust"
        vectorial_model_version = "0.1.0"
        model_function = run_rust_vectorial_model

    vmr = model_function(vmc=vmc, extra_config=extra_config)  # type: ignore
    execution_time_s = time.time() - model_start_time

    evmc = EncodedVMCalculation(
        vmc=vmc,
        evmr=pickle_to_base64(vmr),
        execution_time_s=execution_time_s,
        vectorial_model_backend=vectorial_model_backend,
        vectorial_model_version=vectorial_model_version,
    )

    # save model into db cache
    if vectorial_model_cache_path:
        save_model_to_cache(
            evmc=evmc, vectorial_model_cache_path=vectorial_model_cache_path
        )

    return evmc


def run_vectorial_models(
    vmc_list: List[VectorialModelConfig],
    extra_config: Union[
        PythonModelExtraConfig, FortranModelExtraConfig, RustModelExtraConfig
    ] = PythonModelExtraConfig(print_progress=False),
    parallelism: int = 1,
    vectorial_model_cache_path: Optional[pathlib.Path] = None,
    verbose: bool = False,
) -> List[VMCalculation]:
    """
    Take a list of model configs and run them in parallel, {parallelism} at a time.

    If a cache path is specified, then we only run unique models to avoid trying to
    save the same model to the db at the same time
    """
    if vectorial_model_cache_path is not None:
        # if we are caching the models, run the unique model configs and cache them,
        # then build the results list like normal
        run_unique_models_only(
            vmc_list=vmc_list,
            extra_config=extra_config,
            parallelism=parallelism,
            vectorial_model_cache_path=vectorial_model_cache_path,
            verbose=verbose,
        )

    verbose_print("Caching done, running...", verbose=verbose)

    # The fortran version uses fixed file names for input and output, so running multiple in parallel
    # would clobber each other's input and output files
    if isinstance(extra_config, FortranModelExtraConfig):
        verbose_print("Forcing no parallelism for fortran version!", verbose=verbose)
        parallelism = 1

    pool_start_time = time.time()

    run_vmodel_timed_mappable_func = partial(
        run_vectorial_model_single,
        extra_config=extra_config,
        vectorial_model_cache_path=vectorial_model_cache_path,
    )

    with Pool(parallelism) as vm_pool:
        vmcalc_list = vm_pool.map(run_vmodel_timed_mappable_func, vmc_list)

    pool_end_time = time.time()
    verbose_print(
        f"Total results assembly time: {pool_end_time - pool_start_time} seconds",
        verbose=verbose,
    )

    return [VMCalculation.from_encoded(x) for x in vmcalc_list]


def run_unique_models_only(
    vmc_list: List[VectorialModelConfig],
    vectorial_model_cache_path: pathlib.Path,
    extra_config: Union[
        PythonModelExtraConfig, FortranModelExtraConfig, RustModelExtraConfig
    ] = PythonModelExtraConfig(print_progress=False),
    parallelism: int = 1,
    verbose: bool = False,
) -> None:
    """
    This compares the given vmc_list hashes and runs only the models which are different,
    saving the results into the database at vectorial_model_cache_path - no results are returned directly.
    This allows us to later "run" the full vmc_list and pull the results out of the database
    in the same order.
    """
    # The fortran version uses fixed file names for input and output, so running multiple in parallel
    # would clobber each other's input and output files without some extra temporary-directory work
    # that isn't really worth it
    if isinstance(extra_config, FortranModelExtraConfig):
        verbose_print("Forcing no parallelism for fortran version!", verbose=verbose)
        parallelism = 1

    vmc_hashes = [vmc_to_sha256_digest(x) for x in vmc_list]
    unique_vmc_hashes = set(vmc_hashes)
    reduced_vmc_set = [
        vmc_list[vmc_hashes.index(unique_hash)] for unique_hash in unique_vmc_hashes
    ]
    verbose_print(
        f"Number of unique models to be run and cached: {len(reduced_vmc_set)}",
        verbose=verbose,
    )

    pool_start_time = time.time()

    run_vmodel_timed_mappable_func = partial(
        run_vectorial_model_single,
        extra_config=extra_config,
        vectorial_model_cache_path=vectorial_model_cache_path,
    )

    with Pool(parallelism) as vm_pool:
        # run the results and let them be saved in db
        _ = vm_pool.map(run_vmodel_timed_mappable_func, reduced_vmc_set)

    pool_end_time = time.time()
    verbose_print(
        f"Total run time on unique models: {pool_end_time - pool_start_time} seconds",
        verbose=verbose,
    )



from typing import List

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PositiveInt,
    PositiveFloat,
    field_validator,
)
import astropy.units as u


# Helper classes for supporting astropy's Quantity values with pydantic validation
class TimeQuantity(u.Quantity):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, s):
        s = u.Quantity(s)
        if s.unit is None:
            raise ValueError(f"No units for {s}!")
        if s.unit.physical_type != "time":
            raise ValueError(f"Incorrect unit for time: {s.unit}")
        return s


class AreaQuantity(u.Quantity):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, s):
        s = u.Quantity(s)
        if s.unit is None:
            raise ValueError(f"No units for {s}!")
        if s.unit.physical_type != "area":
            raise ValueError(f"Incorrect unit for area: {s.unit}")
        return s


class RateQuantity(u.Quantity):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, s):
        s = u.Quantity(s)
        if s.unit is None:
            raise ValueError(f"No units for {s}!")
        if s.unit.physical_type != "frequency":
            raise ValueError(f"Incorrect unit for rate: {s.unit}")
        return s


class SpeedQuantity(u.Quantity):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, s):
        s = u.Quantity(s)
        if s.unit is None:
            raise ValueError(f"No units for {s}!")
        if s.unit.physical_type != "speed":
            raise ValueError(f"Incorrect unit for speed: {s.unit}")
        return s


class ProductionTimeVariation(BaseModel):
    model_config = ConfigDict(
        json_encoders={u.Quantity: lambda v: f"{v.value} {v.unit}"}, frozen=True
    )
    time_variation_type: str


class GaussianProductionTimeVariation(ProductionTimeVariation):
    amplitude_per_s: float
    std_dev_hrs: PositiveFloat
    t_max_hrs: float

    @field_validator("time_variation_type")
    def validate_time_variation_type(cls, tvt):
        if tvt.lower() != "gaussian":
            raise ValueError(
                f"Incorrect time variation label {tvt} for a gaussian variation!"
            )
        return tvt

    @property
    def amplitude(self) -> RateQuantity:
        return self.amplitude_per_s / u.s

    @property
    def std_dev(self) -> TimeQuantity:
        return self.std_dev_hrs * u.hour  # type: ignore

    @property
    def t_max(self) -> TimeQuantity:
        return self.t_max_hrs * u.hour  # type: ignore


class SineWaveProductionTimeVariation(ProductionTimeVariation):
    amplitude_per_s: float
    period_hrs: float
    delta_hrs: float

    @field_validator("time_variation_type")
    def validate_time_variation_type(cls, tvt):
        if tvt.lower() != "sine wave":
            raise ValueError(
                f"Incorrect time variation label {tvt} for a sine wave variation!"
            )
        return tvt

    @property
    def amplitude(self) -> RateQuantity:
        return self.amplitude_per_s / u.s

    @property
    def period(self) -> TimeQuantity:
        return self.period_hrs * u.hour  # type: ignore

    @property
    def delta(self) -> TimeQuantity:
        return self.delta_hrs * u.hour  # type: ignore


class SquarePulseProductionTimeVariation(ProductionTimeVariation):
    amplitude_per_s: float
    duration_hrs: PositiveFloat
    t_start_hrs: float

    @field_validator("time_variation_type")
    def validate_time_variation_type(cls, tvt):
        if tvt.lower() != "square pulse":
            raise ValueError(
                f"Incorrect time variation label {tvt} for a square wave variation!"
            )
        return tvt

    @property
    def amplitude(self) -> RateQuantity:
        return self.amplitude_per_s / u.s

    @property
    def duration(self) -> TimeQuantity:
        return self.duration_hrs * u.hour  # type: ignore

    @property
    def t_start(self) -> TimeQuantity:
        return self.t_start_hrs * u.hour  # type: ignore


class BinnedProductionTimeVariation(ProductionTimeVariation):
    q_per_s: List[float]
    times_at_productions_days: List[float]

    @field_validator("time_variation_type")
    def validate_time_variation_type(cls, tvt):
        if tvt.lower() != "binned":
            raise ValueError(
                f"Incorrect time variation label {tvt} for binned production!"
            )
        return tvt

    @property
    def q(self) -> List[RateQuantity]:
        return self.q_per_s / u.s

    @property
    def times_at_productions(self) -> List[TimeQuantity]:
        return self.times_at_productions_days * u.day  # type: ignore


class CometProduction(BaseModel):
    model_config = ConfigDict(
        json_encoders={u.Quantity: lambda v: f"{v.value} {v.unit}"}, frozen=True
    )

    base_q_per_s: PositiveFloat
    time_variation: (
        GaussianProductionTimeVariation
        | SineWaveProductionTimeVariation
        | SquarePulseProductionTimeVariation
        | BinnedProductionTimeVariation
        | None
    ) = None

    @property
    def base_q(self) -> RateQuantity:
        return self.base_q_per_s / u.s


class ParentMolecule(BaseModel):
    model_config = ConfigDict(
        json_encoders={u.Quantity: lambda v: f"{v.value} {v.unit}"}, frozen=True
    )

    tau_d_s: PositiveFloat
    tau_T_s: PositiveFloat
    v_outflow_kms: PositiveFloat
    sigma_cm_sq: PositiveFloat

    @property
    def tau_d(self) -> TimeQuantity:
        return self.tau_d_s * u.s

    @property
    def tau_T(self) -> TimeQuantity:
        return self.tau_T_s * u.s

    @property
    def v_outflow(self) -> SpeedQuantity:
        return self.v_outflow_kms * (u.km / u.s)

    @property
    def sigma(self) -> AreaQuantity:
        return self.sigma_cm_sq * (u.cm**2)


class FragmentMolecule(BaseModel):
    model_config = ConfigDict(
        json_encoders={u.Quantity: lambda v: f"{v.value} {v.unit}"}, frozen=True
    )

    tau_T_s: PositiveFloat
    v_photo_kms: PositiveFloat

    @property
    def tau_T(self) -> TimeQuantity:
        return self.tau_T_s * u.s

    @property
    def v_photo(self) -> SpeedQuantity:
        return self.v_photo_kms * (u.km / u.s)


class VectorialModelGrid(BaseModel):
    model_config = ConfigDict(frozen=True)
    radial_points: PositiveInt
    angular_points: PositiveInt
    radial_substeps: PositiveInt
    parent_destruction_level: PositiveFloat = Field(le=1.0)
    fragment_destruction_level: PositiveFloat = Field(le=1.0)


class VectorialModelConfig(BaseModel):
    model_config = ConfigDict(frozen=True)
    production: CometProduction
    parent: ParentMolecule
    fragment: FragmentMolecule
    grid: VectorialModelGrid



import numpy as np
import sbpy.activity as sba

from dataclasses import dataclass
from scipy.interpolate import PPoly
from astropy.units.quantity import Quantity
from typing import Optional

from pyvectorial_au.model_output.fragment_sputter import FragmentSputterSpherical

"""
Defines VectorialModelResult, a dataclass for holding the result of a vectorial model calculation,
namely, the volume density and fragment sputter.  Functions written to handle VectorialModelResult
don't need to care which model generated the results.
"""


@dataclass
class VectorialModelResult:
    """
    Holds a standardized model result to abstract away the underlying details so that the python,
    fortran, and rust models can be run as backends interchangeably
    """

    """The model calculates the volume density along gridded radial values"""
    volume_density_grid: np.ndarray
    volume_density: np.ndarray

    """ The fragment sputter that produces the volume density above """
    fragment_sputter: FragmentSputterSpherical

    """ Some physical quantities of interest """
    collision_sphere_radius: Quantity
    max_grid_radius: Quantity
    coma_radius: Quantity

    column_density_grid: Optional[np.ndarray] = None
    column_density: Optional[np.ndarray] = None

    volume_density_interpolation: Optional[PPoly] = None
    column_density_interpolation: Optional[PPoly] = None

    coma: Optional[sba.VectorialModel] = None



import os
import pytest
import numpy as np
import astropy.units as u
from pyvectorial_au.pre_model_processing.timedependentproduction import (
    make_time_dependence_function,
)
from pyvectorial_au.model_input.vectorial_model_config_reader import (
    vectorial_model_config_from_yaml,
)


@pytest.fixture
def rootdir():
    return os.path.dirname(os.path.abspath(__file__))


# binned production fixtures
@pytest.fixture
def single_run_binned_yaml(rootdir):
    return os.path.join(rootdir, "input/single_run_binned.yaml")


@pytest.fixture
def vmc_binned(single_run_binned_yaml):
    return vectorial_model_config_from_yaml(single_run_binned_yaml)


# gaussian fixtures
@pytest.fixture
def single_run_gaussian_yaml(rootdir):
    return os.path.join(rootdir, "input/single_run_gaussian.yaml")


@pytest.fixture
def vmc_gaussian(single_run_gaussian_yaml):
    return vectorial_model_config_from_yaml(single_run_gaussian_yaml)


@pytest.fixture
def q_t_gaussian(vmc_gaussian):
    return make_time_dependence_function(vmc_gaussian)


# sine wave fixtures
@pytest.fixture
def single_run_sine_yaml(rootdir):
    return os.path.join(rootdir, "input/single_run_sine.yaml")


@pytest.fixture
def vmc_sine(single_run_sine_yaml):
    return vectorial_model_config_from_yaml(single_run_sine_yaml)


@pytest.fixture
def q_t_sine(vmc_sine):
    return make_time_dependence_function(vmc_sine)


# square pulse fixtures
@pytest.fixture
def single_run_square_pulse_yaml(rootdir):
    return os.path.join(rootdir, "input/single_run_square_pulse.yaml")


@pytest.fixture
def vmc_square_pulse(single_run_square_pulse_yaml):
    return vectorial_model_config_from_yaml(single_run_square_pulse_yaml)


@pytest.fixture
def q_t_square_pulse(vmc_square_pulse):
    return make_time_dependence_function(vmc_square_pulse)


# binned tests
def test_binned_q_first(vmc_binned):
    assert np.isclose(vmc_binned.production.time_variation.q[0], 7.0e25 / u.s)


def test_binned_q_last(vmc_binned):
    assert np.isclose(vmc_binned.production.time_variation.q[-1], 1.0e28 / u.s)


def test_binned_times_first(vmc_binned):
    assert np.isclose(
        vmc_binned.production.time_variation.times_at_productions[0], 35 * u.day  # type: ignore
    )


def test_binned_times_last(vmc_binned):
    assert np.isclose(
        vmc_binned.production.time_variation.times_at_productions[-1], 5 * u.day  # type: ignore
    )


# gaussian tests
def test_gaussian_amplitude(vmc_gaussian):
    assert np.isclose(vmc_gaussian.production.time_variation.amplitude, 3.0e28 / u.s)


def test_gaussian_std_dev(vmc_gaussian):
    assert np.isclose(vmc_gaussian.production.time_variation.std_dev, 20.0 * u.hour)  # type: ignore


def test_gaussian_t_max(vmc_gaussian):
    assert np.isclose(vmc_gaussian.production.time_variation.t_max, 24.0 * u.hour)  # type: ignore


def test_gaussian_q_at_t_max(q_t_gaussian):
    assert np.isclose(q_t_gaussian((24.0 * u.hour).to_value(u.s)), 3.0e28)  # type: ignore


# time for half max is mu +/- sqrt(2 ln(2)) * sigma
def test_gaussian_q_at_half_max_one(q_t_gaussian):
    assert np.isclose(
        q_t_gaussian(((24.0 - np.sqrt(2 * np.log(2)) * 20.0) * u.hour).to_value(u.s)),
        1.5e28,
    )


def test_gaussian_q_at_half_max_two(q_t_gaussian):
    assert np.isclose(
        q_t_gaussian(((24.0 + np.sqrt(2 * np.log(2)) * 20.0) * u.hour).to_value(u.s)),
        1.5e28,
    )


# sine tests
def test_sine_amplitude(vmc_sine):
    assert np.isclose(vmc_sine.production.time_variation.amplitude, 3.0e28 / u.s)


def test_sine_period(vmc_sine):
    assert np.isclose(vmc_sine.production.time_variation.period, 20.0 * u.hour)  # type: ignore


def test_sine_delta(vmc_sine):
    assert np.isclose(vmc_sine.production.time_variation.delta, 3.0 * u.hour)  # type: ignore


def test_sine_q_t_at_delta(q_t_sine):
    assert np.isclose(q_t_sine((-3.0 * u.hour).to_value(u.s)), 0.0)  # type: ignore


def test_sine_q_t_at_first_max_in_past(q_t_sine):
    assert np.isclose(q_t_sine((2.0 * u.hour).to_value(u.s)), 3.0e28)  # type: ignore


def test_sine_q_t_at_first_min_in_past(q_t_sine):
    assert np.isclose(q_t_sine((12.0 * u.hour).to_value(u.s)), -3.0e28)  # type: ignore


# square pulse tests
def test_square_pulse_amplitude(vmc_square_pulse):
    assert np.isclose(
        vmc_square_pulse.production.time_variation.amplitude, 4.0e28 / u.s
    )


def test_square_pulse_duration(vmc_square_pulse):
    assert np.isclose(vmc_square_pulse.production.time_variation.duration, 5.0 * u.hour)  # type: ignore


def test_square_pulse_t_start(vmc_square_pulse):
    assert np.isclose(vmc_square_pulse.production.time_variation.t_start, 10.0 * u.hour)  # type: ignore


def test_q_t_square_pulse_during(q_t_square_pulse):
    assert np.isclose(q_t_square_pulse((7.0 * u.hour).to_value(u.s)), 4.0e28)  # type: ignore


def test_q_t_square_pulse_before(q_t_square_pulse):
    assert np.isclose(q_t_square_pulse((15.0 * u.hour).to_value(u.s)), 0.0)  # type: ignore


def test_q_t_square_pulse_after(q_t_square_pulse):
    assert np.isclose(q_t_square_pulse((0.0 * u.hour).to_value(u.s)), 0.0)  # type: ignore



import numpy as np
import astropy.units as u

from dataclasses import dataclass
from typing import Optional
from pyvectorial_au.model_input.vectorial_model_config import VectorialModelConfig


"""
    Like VectorialModelConfig, HaserParams serves as the standard input data structure, but for Haser models
"""


@dataclass
class HaserParams:
    # total parent production, ~ 1/u.s
    q: Optional[u.Quantity]
    # outflow speed of parents
    v_outflow: Optional[u.Quantity]
    # length scale of parent
    gamma_p: Optional[u.Quantity]
    # length scale of daughter
    gamma_d: Optional[u.Quantity]


def haser_from_vectorial_cd1980(vmc: VectorialModelConfig) -> HaserParams:
    # use relations in Combi & Delsemme 1980 to translate vectorial parameters
    # into roughly equivalent haser parameters

    v_d = np.sqrt(vmc.parent.v_outflow**2 + vmc.fragment.v_photo**2)

    gamma_p = vmc.parent.v_outflow * vmc.parent.tau_T
    gamma_d = v_d * vmc.fragment.tau_T

    delta = np.arctan(vmc.parent.v_outflow / vmc.fragment.v_photo)

    mu = gamma_p / gamma_d
    mu_h = mu * (mu + np.sin(delta)) / (1 + mu * np.sin(delta))

    gamma_d_h = np.sqrt((gamma_d**2 - gamma_p**2) / (mu_h**2 + 1))
    gamma_p_h = mu_h * gamma_d_h

    v_d_h = (v_d * gamma_d_h) / gamma_d

    # TODO: does the paper give a formula for the production, or should we use the same production as went
    # into the vectorial model?  If we fit this new haser and find the best Q, does it match the vectorial?
    # We return None for the production because ... ?
    return HaserParams(q=None, v_outflow=v_d_h, gamma_p=gamma_p_h, gamma_d=gamma_d_h)



import pathlib
from typing import Optional

from sqlalchemy import Engine, NullPool, create_engine
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    Session,
    mapped_column,
    scoped_session,
    sessionmaker,
)

"""
From the sqlalchemy docs:
    When using an Engine with multiple Python processes,
    such as when using os.fork or Python multiprocessing,
    it’s important that the engine is initialized per process.
"""


class VMCachedBase(DeclarativeBase):
    pass


class VMCached(VMCachedBase):
    __tablename__ = "vmcache"
    vmc_hash: Mapped[str] = mapped_column(primary_key=True)
    vmr_b64_enc_zip: Mapped[bytes] = mapped_column(nullable=False)
    execution_time_s: Mapped[float] = mapped_column(nullable=False)
    vectorial_model_backend: Mapped[str] = mapped_column(nullable=False)
    vectorial_model_version: Mapped[str] = mapped_column(nullable=False)


def get_vectorial_model_cache_engine(
    vectorial_model_cache_path: pathlib.Path,
) -> Engine:
    """
    Returns an engine to the db based on the given path
    """

    vm_cache_engine = create_engine(
        f"sqlite:///{vectorial_model_cache_path}", poolclass=NullPool
    )

    return vm_cache_engine


def get_vectorial_model_cache_session(
    vectorial_model_cache_path: pathlib.Path,
    cache_engine: Optional[Engine] = None,
) -> Optional[Session]:
    """
    Returns a db session based on the given Engine, but if none was given we construct
    an engine with the given path.
    If an engine is specified, the path is ignored.
    """

    if not cache_engine:
        cache_engine = get_vectorial_model_cache_engine(
            vectorial_model_cache_path=vectorial_model_cache_path,
        )

    session_factory = sessionmaker(cache_engine)
    cache_session = scoped_session(session_factory)
    return cache_session()



import yaml
import json
import pathlib
import logging as log
from typing import Optional

from pyvectorial_au.model_input.vectorial_model_config import VectorialModelConfig


def vectorial_model_config_from_yaml(
    filepath: pathlib.Path,
) -> Optional[VectorialModelConfig]:
    with open(filepath, "r") as stream:
        try:
            input_yaml = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            log.info("Reading file %s resulted in yaml error: %s", filepath, exc)
            return None

    return VectorialModelConfig(**input_yaml)


def vectorial_model_config_from_json(
    filepath: pathlib.Path,
) -> Optional[VectorialModelConfig]:
    with open(filepath, "r") as f:
        try:
            input_json = json.load(f)
        except json.JSONDecodeError as msg:
            log.info("Reading file %s resulted in json error: %s", filepath, msg)
            return None

    return VectorialModelConfig(**input_json)



import pytest
import astropy.units as u
import numpy as np

from pyvectorial_au.model_input.vectorial_model_config import (
    VectorialModelConfig,
    CometProduction,
    ParentMolecule,
    FragmentMolecule,
    VectorialModelGrid,
)
from pyvectorial_au.pre_model_processing.input_transforms import (
    apply_input_transform,
    VmcTransform,
)


@pytest.fixture
def testing_vmc() -> VectorialModelConfig:
    """
    Return a valid VectorialModelConfig for testing input transformations
    """

    q = CometProduction(base_q_per_s=1e28)
    p = ParentMolecule(
        v_outflow_kms=1.0,
        tau_d_s=100000,
        tau_T_s=93000,
        sigma_cm_sq=3.0e-16,
    )
    f = FragmentMolecule(v_photo_kms=1.05, tau_T_s=100000)
    g = VectorialModelGrid(
        radial_points=50,
        angular_points=50,
        radial_substeps=50,
        parent_destruction_level=0.99,
        fragment_destruction_level=0.95,
    )

    return VectorialModelConfig(production=q, parent=p, fragment=f, grid=g)


def test_fortran_festou_apply(testing_vmc):
    r_h = 2.0 * u.AU  # type: ignore
    r_h_au = r_h.to_value(u.AU)
    rhsq = r_h_au**2

    xfrmed_vmc = apply_input_transform(
        vmc=testing_vmc, r_h=r_h, xfrm=VmcTransform.fortran_festou
    )
    assert xfrmed_vmc is not None

    transformed_input = [
        xfrmed_vmc.parent.tau_d.to_value(u.s),  # type: ignore
        xfrmed_vmc.parent.tau_T.to_value(u.s),  # type: ignore
        xfrmed_vmc.fragment.tau_T.to_value(u.s),  # type: ignore
    ]
    expected_result = [
        testing_vmc.parent.tau_d.to_value(u.s) * rhsq,
        testing_vmc.parent.tau_T.to_value(u.s) * rhsq,
        testing_vmc.fragment.tau_T.to_value(u.s) * rhsq,
    ]

    assert np.allclose(transformed_input, expected_result)  # type: ignore


def test_cochran_schleicher_93_apply(testing_vmc):
    r_h = 2.0 * u.AU  # type: ignore
    r_h_au = r_h.to_value(u.AU)
    rhsq = r_h_au**2
    sqrh = np.sqrt(r_h_au)

    xfrmed_vmc = apply_input_transform(
        testing_vmc, r_h, VmcTransform.cochran_schleicher_93
    )
    assert xfrmed_vmc is not None

    transformed_input = [
        xfrmed_vmc.parent.tau_d.to_value(u.s),  # type: ignore
        xfrmed_vmc.parent.tau_T.to_value(u.s),  # type: ignore
        xfrmed_vmc.fragment.tau_T.to_value(u.s),  # type: ignore
        xfrmed_vmc.parent.v_outflow.to_value(u.km / u.s),  # type: ignore
    ]
    expected_result = [
        testing_vmc.parent.tau_d.to(u.s).value * rhsq,
        testing_vmc.parent.tau_T.to(u.s).value * rhsq,
        testing_vmc.fragment.tau_T.to(u.s).value * rhsq,
        0.85 / sqrh,
    ]

    assert np.allclose(transformed_input, expected_result)  # type: ignore



import zlib
import hashlib
import dill as pickle
import codecs
from typing import Any, TypeAlias

from pyvectorial_au.model_input.vectorial_model_config import VectorialModelConfig

VectorialModelConfigHash: TypeAlias = str


# Serialize an object to a base64 string for storage
def pickle_to_base64(obj: Any) -> str:
    return codecs.encode(pickle.dumps(obj), "base64").decode()


# Restore object from base64 string encoding
def unpickle_from_base64(p_str: str) -> Any:
    return pickle.loads(codecs.decode(p_str.encode(), "base64"))


# return a string of the sha256 hex digest of the given vmc
def vmc_to_sha256_digest(vmc: VectorialModelConfig) -> VectorialModelConfigHash:
    h = hashlib.sha256()
    h.update(vmc.model_dump_json().encode())

    return h.hexdigest()


def compress_vmr_string(vmr: str) -> bytes:
    return zlib.compress(vmr.encode())


def decompress_vmr_string(vmr_compressed: bytes) -> str:
    return zlib.decompress(vmr_compressed).decode()



import logging as log
from typing import Callable, Optional

import astropy.units as u
import numpy as np

from pyvectorial_au.model_input.vectorial_model_config import (
    GaussianProductionTimeVariation,
    SineWaveProductionTimeVariation,
    SquarePulseProductionTimeVariation,
    VectorialModelConfig,
)


def make_time_dependence_function(vmc: VectorialModelConfig) -> Optional[Callable]:
    if isinstance(vmc.production.time_variation, GaussianProductionTimeVariation):
        return make_gaussian_q_t(vmc=vmc)
    elif isinstance(vmc.production.time_variation, SineWaveProductionTimeVariation):
        return make_sine_q_t(vmc=vmc)
    elif isinstance(vmc.production.time_variation, SquarePulseProductionTimeVariation):
        return make_square_pulse_q_t(vmc=vmc)
    else:
        return None


def make_gaussian_q_t(vmc: VectorialModelConfig) -> Callable:
    assert isinstance(vmc.production.time_variation, GaussianProductionTimeVariation)

    """Assembles a gaussian time dependence based on the parameters"""
    amplitude_in_invsecs = vmc.production.time_variation.amplitude.to_value(1 / u.s)  # type: ignore
    std_dev_in_secs = vmc.production.time_variation.std_dev.to_value(u.s)  # type: ignore
    t_max_in_secs = vmc.production.time_variation.t_max.to_value(u.s)  # type: ignore

    log.debug(
        "Building gaussian q_t:\tAmplitude: %s, t_max: %s, std_dev: %s",
        vmc.production.time_variation.amplitude,
        vmc.production.time_variation.t_max,
        vmc.production.time_variation.std_dev,
    )

    def q_t(t):
        return amplitude_in_invsecs * np.e ** -(
            ((t - t_max_in_secs) ** 2) / (2 * std_dev_in_secs**2)  # type: ignore
        )

    return q_t


def make_sine_q_t(vmc: VectorialModelConfig) -> Callable:
    assert isinstance(vmc.production.time_variation, SineWaveProductionTimeVariation)

    """Assembles a sinusoidal time dependence based on the parameters"""
    amplitude_in_invsecs = vmc.production.time_variation.amplitude.to_value(1 / u.s)  # type: ignore
    period_in_secs = vmc.production.time_variation.period.to_value(u.s)  # type: ignore
    delta_in_secs = vmc.production.time_variation.delta.to_value(u.s)  # type: ignore
    const_B = (2.0 * np.pi) / period_in_secs  # type: ignore

    log.debug(
        "Building sinusoidal q_t:\tAmplitude: %s, period: %s, delta: %s",
        vmc.production.time_variation.amplitude,
        vmc.production.time_variation.period,
        vmc.production.time_variation.delta,
    )

    def q_t(t):
        return amplitude_in_invsecs * (np.sin(const_B * (t + delta_in_secs)))

    return q_t


def make_square_pulse_q_t(vmc: VectorialModelConfig) -> Callable:
    assert isinstance(vmc.production.time_variation, SquarePulseProductionTimeVariation)

    """Assembles a square pulse time dependence based on the parameters"""
    t_start_in_secs = vmc.production.time_variation.t_start.to_value(u.s)  # type: ignore
    tend_in_secs = (  # type: ignore
        vmc.production.time_variation.t_start - vmc.production.time_variation.duration
    ).to_value(  # type: ignore
        u.s
    )
    amplitude_in_invsecs = vmc.production.time_variation.amplitude.to_value(1 / u.s)  # type: ignore

    log.debug(
        "Building square pulse q_t:\tAmplitude: %s, t_start: %s, duration: %s",
        vmc.production.time_variation.amplitude,
        vmc.production.time_variation.t_start,
        vmc.production.time_variation.duration,
    )

    def q_t(t):
        # Comparisons seem backward because of our weird time system
        if t < t_start_in_secs and t > tend_in_secs:
            return amplitude_in_invsecs
        else:
            return 0.0

    return q_t



import os
import pytest
import astropy.units as u
import numpy as np
from pyvectorial_au.backends.fortran_version import vmr_from_fortran_output


@pytest.fixture
def rootdir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def fortran_fort16_file(rootdir):
    return os.path.join(rootdir, "input/fort.16")


@pytest.fixture
def vmc_from_fortran(fortran_fort16_file):
    return vmr_from_fortran_output(fortran_fort16_file, read_sputter=True)


def test_fortran_collision_sphere(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.collision_sphere_radius, 0.9e7 * u.cm)


def test_fortran_max_grid_radius(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.max_grid_radius, 1.061e6 * u.km)  # type: ignore


def test_fortran_coma_radius(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.coma_radius, 7.259e5 * u.km)  # type: ignore


# def test_num_fragments_theory(vmc_from_fortran):
#     assert np.isclose(vmc_from_fortran.num_fragments_theory, 0.545e33)
#
#
# def test_num_fragments_grid(vmc_from_fortran):
#     assert np.isclose(vmc_from_fortran.num_fragments_grid, 0.533e33)


def test_volume_density_grid_first(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.volume_density_grid[0], 0.2e3 * u.km)  # type: ignore


def test_volume_density_grid_last(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.volume_density_grid[-1], 0.10e7 * u.km)  # type: ignore


def test_volume_density_first(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.volume_density[0], 0.34e5 / u.cm**3)


def test_volume_density_last(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.volume_density[-1], 0.15e-3 / u.cm**3)


def test_column_density_grid_first(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.column_density_grid[0], 0.20322e3 * u.km)  # type: ignore


def test_column_density_grid_last(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.column_density_grid[-1], 7.21038e5 * u.km)  # type: ignore


def test_column_density_first(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.column_density[0], 7.73995e12 / u.cm**2)


def test_column_density_last(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.column_density[-1], 4.88174e8 / u.cm**2)


def test_fragment_sputter_r_first(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.fragment_sputter.rs[0], 1048352.88 * u.km)  # type: ignore


def test_fragment_sputter_r_last(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.fragment_sputter.rs[-1], 203.169159 * u.km)  # type: ignore


def test_fragment_sputter_theta_first(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.fragment_sputter.thetas[0], 6.04152456e-2)


def test_fragment_sputter_theta_last(vmc_from_fortran):
    assert np.isclose(vmc_from_fortran.fragment_sputter.thetas[-1], 3.08117747)


def test_fragment_sputter_density_first(vmc_from_fortran):
    assert np.isclose(
        vmc_from_fortran.fragment_sputter.fragment_density[0], 6.28462021e-6 / u.cm**3
    )


def test_fragment_sputter_density_last(vmc_from_fortran):
    assert np.isclose(
        vmc_from_fortran.fragment_sputter.fragment_density[-1], 3.87345695 / u.cm**3
    )



import os
import pytest
import astropy.units as u
from pyvectorial_au.model_input.vectorial_model_config_reader import (
    vectorial_model_config_from_yaml,
)


@pytest.fixture
def rootdir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def single_run_yaml(rootdir):
    return os.path.join(rootdir, "input/single_run.yaml")


@pytest.fixture
def vmc_set_single(single_run_yaml):
    return vectorial_model_config_from_yaml(single_run_yaml)


def test_single_run_base_q(vmc_set_single):
    assert vmc_set_single.production.base_q == (1.0e28 / u.s)


def test_single_run_parent_outflow(vmc_set_single):
    assert vmc_set_single.parent.v_outflow == (0.85 * u.km / u.s)  # type: ignore


def test_single_run_parent_tau_T(vmc_set_single):
    assert vmc_set_single.parent.tau_T == (45000 * u.s)


def test_single_run_parent_tau_d(vmc_set_single):
    assert vmc_set_single.parent.tau_d == (50000 * u.s)


def test_single_run_parent_sigma(vmc_set_single):
    assert vmc_set_single.parent.sigma == (3.0e-16 * u.cm**2)  # type: ignore


def test_single_run_fragment_v_photo(vmc_set_single):
    assert vmc_set_single.fragment.v_photo == (1.05 * u.km / u.s)  # type: ignore


def test_single_run_fragment_tau_T(vmc_set_single):
    assert vmc_set_single.fragment.tau_T == (129000 * u.s)


def test_single_run_grid_radial_points(vmc_set_single):
    assert vmc_set_single.grid.radial_points == 50


def test_single_run_grid_angular_points(vmc_set_single):
    assert vmc_set_single.grid.angular_points == 30


def test_single_run_grid_radial_substeps(vmc_set_single):
    assert vmc_set_single.grid.radial_substeps == 40



import logging as log
from typing import Tuple

import numpy as np
import scipy.integrate
import astropy.units as u
import sbpy.activity as sba

from pyvectorial_au.model_output.vectorial_model_result import VectorialModelResult


class UncenteredRectangularAperture(sba.Aperture):
    """Rectangular aperture projected at the distance of the target.

    Parameters
    ----------
    shape : `~astropy.units.Quantity`
        A four-element `~astropy.units.Quantity` of angular or
        projected linear size for the width and height of the
        aperture.  The order is not significant.

    """

    def __init__(self, shape):
        if len(shape) != 4:
            raise ValueError("shape must be 4 elements")
        super().__init__(shape)

    def __repr__(self):
        return f"<UncenteredRectangularAperture: coordinates {self.dim[0].value},{self.dim[1].value}×{self.dim[2].value},{self.dim[3].value}>"

    @property
    def shape(self):
        """Rectangle dimensions."""
        return self.dim

    def coma_equivalent_radius(self):
        # this function only makes sense when the aperture is centered
        return 0

    coma_equivalent_radius.__doc__ = sba.Aperture.coma_equivalent_radius.__doc__


def total_number_in_aperture(
    vmr: VectorialModelResult, ap: sba.Aperture, epsabs=1.49e-8
) -> Tuple[float, float]:
    """
    Integrate column density over an aperture, returning tuple of (result, err)
    """

    if vmr.column_density_interpolation is None:
        log.debug(
            "Attempting total_number_in_aperture but column density interpolation is missing!"
        )
        return (0.0, 0.0)

    if isinstance(ap, (sba.CircularAperture, sba.AnnularAperture)):
        if isinstance(ap, sba.CircularAperture):
            limits = (0, ap.radius.to_value(u.m))
        else:
            limits = ap.shape.to_value(u.m)

        # integrate in polar coordinates
        def f_circular(rho):
            """Column density integration in polar coordinates.

            rho in m, column_density in m**-2

            """
            return rho * vmr.column_density_interpolation(rho)  # type: ignore

        N, err = scipy.integrate.quad(f_circular, *limits, epsabs=epsabs)
        N *= 2 * np.pi
        err *= 2 * np.pi
    elif isinstance(ap, sba.RectangularAperture):
        shape = ap.shape.to_value(u.m)

        def f_rectangular(rho, _):
            """Column density integration in polar coordinates.

            rho in m, column_density in m**-2

            th is ignored (azimuthal symmetry)

            """
            return rho * vmr.column_density_interpolation(rho)  # type: ignore

        # first "octant"; rho1 and rho2 are the limits of the
        # integration
        def rho1(_):
            "Lower limit"
            return 0

        def rho2_1(th):
            "Upper limit (a line)"
            return shape[0] / 2 / np.cos(th)

        th = np.arctan(shape[1] / shape[0])
        N1, err1 = scipy.integrate.dblquad(  # type: ignore
            f_rectangular, 0, th, rho1, rho2_1, epsabs=epsabs
        )

        # second "octant"
        def rho2_2(th):
            "Upper limit (a line)"
            return shape[1] / 2 / np.cos(th)

        th = np.arctan(shape[0] / shape[1])
        N2, err2 = scipy.integrate.dblquad(  # type: ignore
            f_rectangular, 0, th, rho1, rho2_2, epsabs=epsabs
        )

        # N1 + N2 constitute 1/4th of the rectangle
        N = 4 * (N1 + N2)
        err = 4 * (err1 + err2)
    elif isinstance(ap, sba.GaussianAperture):
        # integrate in polar coordinates
        def f_gauss(rho, sigma):
            """Column density integration in polar coordinates.

            rho and sigma in m, column_density in m**-2

            """
            return (
                rho
                * np.exp(-(rho**2) / sigma**2 / 2)
                * vmr.column_density_interpolation(rho)  # type: ignore
            )

        sigma = ap.sigma.to_value(u.m)
        N, err = scipy.integrate.quad(f_gauss, 0, np.inf, args=(sigma,), epsabs=epsabs)
        N *= 2 * np.pi
        err *= 2 * np.pi
    elif isinstance(ap, UncenteredRectangularAperture):
        shape = ap.shape.to_value(u.m)

        def f_uncentered(x, y):
            rho = np.sqrt(x**2 + y**2)
            return vmr.column_density_interpolation(rho)  # type: ignore

        # shape = (x1, y1, x2, y2)
        N, err = scipy.integrate.dblquad(  # type: ignore
            f_uncentered, shape[0], shape[2], shape[1], shape[3], epsabs=epsabs
        )
    else:
        N, err = 0, 0
        log.debug("Aperture type %s not handled in total_number_in_aperture!", type(ap))

    return N, err



from dataclasses import dataclass
from typing import TypeAlias

# from functools import partial
# from multiprocessing import Pool

# import pandas as pd
# import dill
# from pydantic import TypeAdapter

from pyvectorial_au.encoding.encoding_and_hashing import unpickle_from_base64
from pyvectorial_au.model_input.vectorial_model_config import VectorialModelConfig
from pyvectorial_au.model_output.vectorial_model_result import VectorialModelResult

EncodedVectorialModelResult: TypeAlias = str


@dataclass
class EncodedVMCalculation:
    """
    Uses a pickled VectorialModelResult (using the dill library) because python multiprocessing wants
    to pickle return values to send them back to the main calling process.  The coma can't be
    pickled by the stock python pickler so we have to encode it ourselves with dill. This allows us
    to return this data structure from a job started in parallel, where returning VMCalculation directly
    would fail.
    """

    vmc: VectorialModelConfig
    evmr: EncodedVectorialModelResult
    execution_time_s: float
    vectorial_model_backend: str
    vectorial_model_version: str


@dataclass
class VMCalculation:
    vmc: VectorialModelConfig
    vmr: VectorialModelResult
    execution_time_s: float
    vectorial_model_backend: str
    vectorial_model_version: str

    @classmethod
    def from_encoded(cls, evmc: EncodedVMCalculation):
        return cls(
            vmc=evmc.vmc,
            vmr=unpickle_from_base64(evmc.evmr),
            execution_time_s=evmc.execution_time_s,
            vectorial_model_backend=evmc.vectorial_model_backend,
            vectorial_model_version=evmc.vectorial_model_version,
        )


# def store_vmcalculation_list(
#     vmcalc_list: List[VMCalculation], out_file: pathlib.Path
# ) -> None:
#     vmcalc_list_pickle = dill.dumps(vmcalc_list)
#     with open(out_file, "wb") as f:
#         f.write(vmcalc_list_pickle)
#
#
# def load_vmcalculation_list(in_file: pathlib.Path) -> List[VMCalculation]:
#     with open(in_file, "rb") as f:
#         pstring = f.read()
#
#     return dill.loads(pstring)
#
#
# def vmcalc_list_to_dataframe(vmcalc_list: List[VMCalculation]):
#     df_list = []
#
#     for vmcalc in vmcalc_list:
#         # flatten the nested model_dump dictionary into column names like production.base_q_per_s with json_normalize()
#         vmc_df = pd.json_normalize(vmcalc.vmc.model_dump())
#         vmr_etc_df = pd.DataFrame(
#             [
#                 {
#                     "vmr_base64": pickle_to_base64(vmcalc.vmr),
#                     "execution_time_s": vmcalc.execution_time_s,
#                     "vectorial_model_backend": vmcalc.vectorial_model_backend,
#                     "vectorial_model_version": vmcalc.vectorial_model_version,
#                     "vmc_sha256_digest": vmc_to_sha256_digest(vmcalc.vmc),
#                 }
#             ]
#         )
#         this_df = pd.concat([vmc_df, vmr_etc_df], axis=1)
#         df_list.append(this_df)
#
#     return pd.concat(df_list)


# def dataframe_to_vmcalc_list(
#     df: pd.DataFrame, column_name_separator: str = "."
# ) -> List[VMCalculation]:
#     json_dict_list: List[dict] = []
#
#     # builds a list of dictionaries that each describe a row of the given dataframe.
#     # takes column names like production.base_q_per_s and puts them in dict keys json_dict_list['production']['base_q_per_s']
#     # by splitting on the column name separator
#     for _, row in df.iterrows():
#         row_dict = {}
#         for column_name, value in row.items():
#             assert isinstance(column_name, str)
#             keys = column_name.split(column_name_separator)
#
#             cur_dict = row_dict
#             for j, key in enumerate(keys):
#                 if j == len(keys) - 1:
#                     cur_dict[key] = value
#                 else:
#                     if key not in cur_dict.keys():
#                         cur_dict[key] = {}
#                     cur_dict = cur_dict[key]
#
#         json_dict_list.append(row_dict)
#
#     # if we have no time variation, the json value is a NaN float value instead of None, so fix that here
#     for jd in json_dict_list:
#         if pd.isna(jd["production"]["time_variation"]):
#             jd["production"]["time_variation"] = None
#
#     ta = TypeAdapter(VectorialModelConfig)
#     vmcl = [
#         VMCalculation(
#             vmc=ta.validate_python(jd),
#             vmr=unpickle_from_base64(jd["vmr_base64"]),
#             execution_time_s=jd["execution_time_s"],
#             vectorial_model_backend=jd["vectorial_model_backend"],
#             vectorial_model_version=jd["vectorial_model_version"],
#         )
#         for jd in json_dict_list
#     ]
#
#     return vmcl



from dataclasses import dataclass
import numpy as np


# Dataclasses for the two-dimensional fragment sputter information obtained from the model
@dataclass
class FragmentSputterSpherical:
    """
    This holds the direct output of the model's fragment sputter in terms of (r, theta, density)
    Theta here is the polar angle in a spherical coordinate system, and the density is azimuthally symmetric
    All of the results are then 0 <= theta <= pi, occupying a half-plane
    """

    rs: np.ndarray
    thetas: np.ndarray
    fragment_density: np.ndarray


@dataclass
class FragmentSputterPolar:
    """
    If we want to expand the fragment density to the whole plane, we use this class to
    remove ambiguity of the domain of theta: 0 <= theta <= 2*pi
    We can then call mirror_fragment_sputter() to fill in the fragment_density to include
    the missing pi <= theta <= 2*pi section of the fragment sputter to have a full-plane
    representation
    """

    rs: np.ndarray
    thetas: np.ndarray
    fragment_density: np.ndarray


@dataclass
class FragmentSputterCartesian:
    """
    Representation of the fragment sputter in (x, y, density).  If we take results from
    the FragmentSputterSpherical, then only the positive x-axis half of the plane is included.
    Like FragmentSputterPolar, we can call mirror_fragment_sputter() to fill in the negative x-axis
    to obtain a whole plane view of the fragment density.
    """

    xs: np.ndarray
    ys: np.ndarray
    fragment_density: np.ndarray



import os
import re
import subprocess
import logging as log
import pathlib
import importlib

import numpy as np
import astropy.units as u
from astropy.units.quantity import Quantity
from itertools import islice
from contextlib import redirect_stdout
from dataclasses import dataclass
from pyvectorial_au.post_model_processing.column_density_abel import (
    column_density_from_abel,
)
from pyvectorial_au.post_model_processing.interpolation import (
    interpolate_column_density,
    interpolate_volume_density,
)

from pyvectorial_au.model_input.vectorial_model_config import VectorialModelConfig
from pyvectorial_au.model_output.vectorial_model_result import (
    VectorialModelResult,
    FragmentSputterSpherical,
)

"""
    For interfacing with the fortran version of the vectorial model, written by Festou, early 1980s

    Takes a VectorialModelConfig and fills a VectorialModelResult based on its calculations
"""


@dataclass
class FortranModelExtraConfig:
    fortran_input_filename: pathlib.Path
    fortran_output_filename: pathlib.Path
    r_h: Quantity
    read_sputter: bool = True
    bin_path: pathlib.Path = importlib.resources.files(  # type: ignore
        package="pyvectorial_au"
    ) / pathlib.Path("bin/fvm_m1")


def run_fortran_vectorial_model(
    vmc: VectorialModelConfig, extra_config: FortranModelExtraConfig
) -> VectorialModelResult:
    """
    Given a valid FortranModelExtraConfig, runs the vectorial model fortran binary it by sending it the correct keystrokes
    """

    # TODO: document that fortran version will transform the input parameters for heliocentric distance, extra_config.r_h
    # TODO: We should probably fix r_h passed to fortran at 1.0 AU and insist that the user scale the VectorialModelConfig parameters themselves

    write_fortran_input_file(vmc, extra_config)

    log.info("Running fortran version at %s ...", extra_config.bin_path)

    # my copy of Festou's vm.f consumes 14 enters before the calculation
    enter_key_string = "\n" * 14
    p1 = subprocess.Popen(["echo", enter_key_string], stdout=subprocess.PIPE)
    p2 = subprocess.run(
        f"{extra_config.bin_path}", stdin=p1.stdout, stdout=open(os.devnull, "wb")
    )

    log.info("fortran run complete, return code %s", p2.returncode)

    vmr = vmr_from_fortran_output(
        extra_config.fortran_output_filename, read_sputter=extra_config.read_sputter
    )
    interpolate_volume_density(vmr)
    column_density_from_abel(vmr)
    interpolate_column_density(vmr)
    return vmr


def vmr_from_fortran_output(
    fort16_file: pathlib.Path, read_sputter: bool
) -> VectorialModelResult:
    """
    Takes the output of the fortran code in fort16_file and returns VectorialModelResult
    """

    log.debug("Attempting to extract VectorialModelConfig from %s ... ", fort16_file)

    # Volume density is on line 15 - 27
    fort16_voldens = range(14, 27)
    # Column density is on line 53 - 70
    fort16_coldens = range(52, 70)
    # sputter array starts at line 177 through to the end
    fort16_sputter = 176

    vdg = []
    vd = []
    cdg = []
    cd = []
    sputter = []
    with open(fort16_file) as in_file:
        fort_header = list(islice(in_file, 30))
        in_file.seek(0)
        for i, line in enumerate(in_file):
            if i in fort16_voldens:
                vals = [float(x) for x in line.split()]
                vdg.extend(vals[0::2])
                vd.extend(vals[1::2])
            if i in fort16_coldens:
                vals = [float(x) for x in line.split()]
                cdg.extend(vals[0::2])
                cd.extend(vals[1::2])
            if i >= fort16_sputter and read_sputter:
                vals = [float(x) for x in line.split()]
                sputter.append(vals)

    mgr_line = fort_header[7]
    mgr = float(re.search("DIM=(.*) KM", mgr_line).group(1)) * u.km  # type: ignore
    cs_line = fort_header[10]
    csphere = float(re.search("is: (.*) cm", cs_line).group(1)) * u.cm  # type: ignore
    cr_line = fort_header[7]
    cr = float(re.search(r"RCOMA=(.*)\(KM\)", cr_line).group(1)) * u.km  # type: ignore

    # fortran outputs are in these units
    vdg *= u.km  # type: ignore
    cdg *= u.km  # type: ignore
    vd *= 1 / u.cm**3  # type: ignore
    cd *= 1 / u.cm**2  # type: ignore

    sputter = np.array(sputter).astype(float)
    rs = sputter[:, 0] * u.km
    thetas = sputter[:, 1]
    fragment_density = sputter[:, 2] / u.cm**3  # type: ignore
    fs = FragmentSputterSpherical(
        rs=rs, thetas=thetas, fragment_density=fragment_density
    )

    log.debug("VectorialModelResult extraction complete.")

    return VectorialModelResult(
        volume_density_grid=vdg,
        volume_density=vd,
        fragment_sputter=fs,
        collision_sphere_radius=csphere,
        max_grid_radius=mgr,
        coma_radius=cr,
        column_density_grid=cdg,
        column_density=cd,
    )


def fragment_theory_count_from_fortran_output(
    fortran_output_filename: pathlib.Path,
) -> float:
    with open(fortran_output_filename) as in_file:
        fort_header = list(islice(in_file, 30))
    nft_line = fort_header[29]
    nft = float(re.search("IS  (.*)  TOTAL", nft_line).group(1))  # type: ignore

    return nft


def fragment_grid_count_from_fortran_output(
    fortran_output_filename: pathlib.Path,
) -> float:
    with open(fortran_output_filename) as in_file:
        fort_header = list(islice(in_file, 30))
    nfg_line = fort_header[28]
    nfg = float(re.search("COMA: (.*)$", nfg_line).group(1))  # type: ignore

    return nfg


def write_fortran_input_file(
    vmc: VectorialModelConfig, ec: FortranModelExtraConfig
) -> None:
    """
    Takes a VectorialModelConfig and produces a fortran input file,
    as long as the production is steady
    """

    if vmc.production.time_variation is not None:
        print(
            "Only steady production is currently supported for producing fortran input files! Running steady production model instead!"
        )

    log.debug(
        "Writing input config file %s to feed fortran vectorial model...",
        ec.fortran_input_filename,
    )

    # default values that do not affect volume or column density, only aperture brightness, which we don't use
    # small delta stos the IUE aperture sizes hard-coded into the fortran version from being too large
    delta = 0.01
    g_factor = 2.33e-4
    comet_name = "Default comet"
    fragment_name = "Default fragment"

    # TODO: binned time production should also be handled here, but fortran only supports 20 bins of time division
    #   so we would have to check if the input had 20 or less time bins first
    # TODO: parent and fragment destruction levels are hard-coded to the defaults of the python model in sbpy so that
    #   the calculations will match

    with open(ec.fortran_input_filename, "w") as out_file:
        with redirect_stdout(out_file):
            print(f"{comet_name}")
            print(f"{ec.r_h.to(u.AU).value}  {delta}")
            # length of production array: only base production rate for 120 days
            print("1")
            print(f"{vmc.production.base_q.to_value(1/u.s)}  120.0")
            # fill in dummy values for the rest of the array
            for _ in range(19):
                print("0.0 121")
            # parent info - speed, total & photo lifetime, destruction level
            print(f"{vmc.parent.v_outflow.to_value(u.km/u.s)}")
            print(f"{vmc.parent.tau_T.to_value(u.s)}")
            print(f"{vmc.parent.tau_d.to_value(u.s)}")
            print(f"{vmc.grid.parent_destruction_level*100}")
            # fragment info - gfactor, speed, total lifetime, destruction level
            print(f"{fragment_name}")
            print(f"{g_factor}")
            print(f"{vmc.fragment.v_photo.to_value(u.km/u.s)}")
            print(f"{vmc.fragment.tau_T.to_value(u.s)}")
            print(f"{vmc.grid.fragment_destruction_level*100}")
            # Custom aperture size, unused for our purposes so these are dummy values
            print("  1.3       3.6 ")
            # print("  0.000001       0.00001")

    log.debug("Writing fortran input file complete.")



import os
import pytest
import astropy.units as u
import numpy as np
from pyvectorial_au.backends.rust_version import vmc_from_rust_output, vmr_from_rust_output


@pytest.fixture
def rootdir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def rust_output_file(rootdir):
    return os.path.join(rootdir, "input/rust_out.txt")


@pytest.fixture
def vmc_from_rust(rust_output_file):
    return vmc_from_rust_output(rust_output_file)


@pytest.fixture
def vmr_from_rust(rust_output_file):
    return vmr_from_rust_output(
        rust_output_file, vmc_from_rust_output(rust_output_file)
    )


def test_rust_collision_sphere(vmr_from_rust):
    assert np.isclose(vmr_from_rust.collision_sphere_radius, 8.82353e1 * u.km)  # type: ignore


def test_rust_max_grid_radius(vmr_from_rust):
    assert np.isclose(vmr_from_rust.max_grid_radius, 1.87195e6 * u.km)  # type: ignore


def test_rust_coma_radius(vmr_from_rust):
    assert np.isclose(vmr_from_rust.coma_radius, 5.58084e5 * u.km)  # type: ignore


def test_volume_density_grid_first(vmr_from_rust):
    assert np.isclose(vmr_from_rust.volume_density_grid[0], 1.0 * u.km)  # type: ignore


def test_volume_density_grid_last(vmr_from_rust):
    assert np.isclose(vmr_from_rust.volume_density_grid[-1], 1.87195e6 * u.km)  # type: ignore


def test_volume_density_first(vmr_from_rust):
    # assert np.isclose(vmr_from_rust.volume_density[0].to_value(1 / u.m**3), (1.62622935 / u.m**3).to_value(1 / u.m**3))  # type: ignore
    assert np.isclose(vmr_from_rust.volume_density[0], 1.62622935e13 / u.m**3)  # type: ignore


def test_volume_density_last(vmr_from_rust):
    assert np.isclose(vmr_from_rust.volume_density[-1], 7.213311e1 / u.m**3)  # type: ignore


def test_fragment_sputter_r_first(vmr_from_rust):
    assert np.isclose(vmr_from_rust.fragment_sputter.rs[0], 1.0 * u.km)  # type: ignore


def test_fragment_sputter_r_last(vmr_from_rust):
    assert np.isclose(vmr_from_rust.fragment_sputter.rs[-1], 1.8719500607868414372205734e6 * u.km)  # type: ignore


def test_fragment_sputter_theta_first(vmr_from_rust):
    assert np.isclose(
        vmr_from_rust.fragment_sputter.thetas[0], 1.9634954084936206974987272e-2
    )


def test_fragment_sputter_theta_last(vmr_from_rust):
    assert np.isclose(
        vmr_from_rust.fragment_sputter.thetas[-1], 3.1219576995048567980006737e0
    )


def test_fragment_sputter_density_first(vmr_from_rust):
    assert np.isclose(
        vmr_from_rust.fragment_sputter.fragment_density[0], 5.0652281546602539062500000e12 / u.m**3  # type: ignore
    )


def test_fragment_sputter_density_last(vmr_from_rust):
    assert np.isclose(
        vmr_from_rust.fragment_sputter.fragment_density[-1], 2.3438452027657809202163594e-24 / u.m**3  # type: ignore
    )



import os
import re
import subprocess
import logging as log
import pathlib
import yaml
import importlib
import tempfile
from itertools import islice
from dataclasses import dataclass
from typing import List, Optional

import numpy as np
import astropy.units as u
from pyvectorial_au.post_model_processing.column_density_abel import (
    column_density_from_abel,
)
from pyvectorial_au.post_model_processing.interpolation import (
    interpolate_column_density,
    interpolate_volume_density,
)

from pyvectorial_au.model_input.vectorial_model_config import (
    VectorialModelConfig,
    CometProduction,
    ParentMolecule,
    FragmentMolecule,
    VectorialModelGrid,
)
from pyvectorial_au.model_output.vectorial_model_result import (
    VectorialModelResult,
    FragmentSputterSpherical,
)


"""
    For interfacing with the rust version of the vectorial model
"""


@dataclass
class RustModelExtraConfig:
    bin_path: pathlib.Path = importlib.resources.files(  # type: ignore
        package="pyvectorial_au"
    ) / pathlib.Path("bin/rust_vect_m1")


def run_rust_vectorial_model(
    vmc: VectorialModelConfig, extra_config: RustModelExtraConfig
) -> VectorialModelResult:
    """
    Given path to rust vmodel binary, runs the given model configuration
    """

    # create the temp files and close them to allow the rust executable sole access
    rust_input = tempfile.NamedTemporaryFile(delete=False)
    rust_input.close()
    rust_output = tempfile.NamedTemporaryFile(delete=False)
    rust_output.close()

    rust_input_filename = pathlib.Path(rust_input.name)
    rust_output_filename = pathlib.Path(rust_output.name)

    log.debug(
        "Writing input config file %s to feed rust vectorial model...",
        rust_input_filename,
    )
    write_rust_input_file(vmc, rust_input_filename)

    log.info("Running rust version at %s ...", extra_config.bin_path)
    p1 = subprocess.run(
        args=[
            str(extra_config.bin_path),
            str(rust_input_filename),
            str(rust_output_filename),
        ],
        stdout=open(os.devnull, "wb"),
    )
    log.info("rust vmodel run complete, return code %s", p1.returncode)

    vmr = vmr_from_rust_output(rust_output_filename, vmc)

    # clean up the temp files
    os.unlink(rust_input_filename)
    os.unlink(rust_output_filename)

    interpolate_volume_density(vmr)
    column_density_from_abel(vmr)
    interpolate_column_density(vmr)
    return vmr


def get_rust_header(rust_output_filename: pathlib.Path) -> List[str]:
    with open(rust_output_filename, "r") as f:
        header = list(islice(f, 14))
    return header


def vmc_from_rust_output(rust_output_filename: pathlib.Path) -> VectorialModelConfig:
    header = get_rust_header(rust_output_filename=rust_output_filename)

    base_q = float(re.search(r"Q: ([0-9]\.[0-9]+e[+-]?[0-9]+) mol/s", header[2]).group(1)) / u.s  # type: ignore
    p_tau_d = (
        float(re.search(r"Parent dissociative lifetime: ([0-9]\.[0-9]*e[+-]?[0-9]+) s", header[2]).group(1))  # type: ignore
        * u.s
    )
    p_tau_T = (
        float(re.search(r"Parent total lifetime: ([0-9]\.[0-9]*e[+-]?[0-9]+) s", header[2]).group(1)) * u.s  # type: ignore
    )
    v_outflow = float(
        re.search(r"Parent outflow velocity: ([0-9]\.[0-9]*e[+-]?[0-9]+) m/s", header[2]).group(1)  # type: ignore
    ) * (u.m / u.s)
    sigma = (
        float(re.search(r"Parent cross sectional area: ([0-9]\.[0-9]*e[+-]?[0-9]+) m\^2", header[3]).group(1))  # type: ignore
        * u.m**2  # type: ignore
    )
    f_tau_T = (
        float(re.search(r"Fragment total lifetime: ([0-9]\.[0-9]*e[+-]?[0-9]+) s", header[3]).group(1)) * u.s  # type: ignore
    )
    v_photo = (
        float(re.search(r"Fragment velocity: ([0-9]\.[0-9]*e[+-]?[0-9]+) m/s", header[3]).group(1)) * u.m / u.s  # type: ignore
    )
    radial_points = int(re.search(r"Radial grid size: ([0-9]+)", header[0]).group(1))  # type: ignore
    angular_points = int(re.search(r"Angular grid size: ([0-9]+)", header[0]).group(1))  # type: ignore
    radial_substeps = int(re.search(r"Radial substeps: ([0-9]+)", header[0]).group(1))  # type: ignore
    parent_destruction_level = (
        float(re.search(r"Parent destruction level: ([0-9]+\.[0-9]+)%", header[4]).group(1)) / 100.0  # type: ignore
    )
    fragment_destruction_level = (
        float(re.search(r"Fragment destruction level: ([0-9]+\.[0-9]+)%", header[4]).group(1))  # type: ignore
        / 100.0
    )

    return VectorialModelConfig(
        production=CometProduction(base_q_per_s=base_q.to_value(1 / u.s)),
        parent=ParentMolecule(
            tau_d_s=p_tau_d.to_value(u.s),
            tau_T_s=p_tau_T.to_value(u.s),
            v_outflow_kms=v_outflow.to_value(u.km / u.s),
            sigma_cm_sq=sigma.to_value(u.cm**2),
        ),
        fragment=FragmentMolecule(
            v_photo_kms=v_photo.to_value(u.km / u.s), tau_T_s=f_tau_T.to_value(u.s)
        ),
        grid=VectorialModelGrid(
            radial_points=radial_points,
            angular_points=angular_points,
            radial_substeps=radial_substeps,
            parent_destruction_level=parent_destruction_level,
            fragment_destruction_level=fragment_destruction_level,
        ),
    )


def vmr_from_rust_output(
    rust_output_filename: pathlib.Path, vmc: Optional[VectorialModelConfig]
) -> VectorialModelResult:
    """
    Takes the output of the rust model in rust_output_filename and returns VectorialModelResult
    """

    if vmc is None:
        vmc = vmc_from_rust_output(rust_output_filename=rust_output_filename)

    # volume density starts at line 14 (counting from 1, not from zero), and has vmc.grid.radial_points entries
    volume_density_start_line_number = 13
    volume_density_stop_line_number = (
        volume_density_start_line_number + vmc.grid.radial_points
    )
    volume_density_lines = range(
        volume_density_start_line_number, volume_density_stop_line_number
    )

    # The +1 skips the line of "r, theta, fragment density"
    sputter_starts_at = volume_density_stop_line_number + 1

    volume_density_grid = []
    volume_density = []
    fragment_sputter_list = []

    with open(rust_output_filename) as f:
        header = list(islice(f, 12))
        f.seek(0)
        for i, line in enumerate(f):
            if i in volume_density_lines:
                vals = [float(x) for x in line.split(",")]
                volume_density_grid.extend(vals[0::2])
                volume_density.extend(vals[1::2])
            if i >= sputter_starts_at:
                vals = [float(x) for x in line.split(",")]
                fragment_sputter_list.append(vals)

    # rust outputs are in these units
    volume_density_grid *= u.m  # type: ignore
    volume_density *= 1 / u.m**3  # type: ignore

    fragment_sputter_array = np.array(fragment_sputter_list).astype(float)
    rs = fragment_sputter_array[:, 0] * u.km
    thetas = fragment_sputter_array[:, 1]
    fragment_density = fragment_sputter_array[:, 2] / u.m**3  # type: ignore
    fragment_sputter = FragmentSputterSpherical(
        rs=rs, thetas=thetas, fragment_density=fragment_density
    )

    collision_sphere_radius = (  # type: ignore
        float(re.search("Collision sphere radius: (.*) km", header[6]).group(1)) * u.km  # type: ignore
    )
    coma_radius = float(re.search("Coma radius: (.*) km", header[7]).group(1)) * u.km  # type: ignore
    max_grid_radius = (  # type: ignore
        float(re.search("Max grid extent: (.*) km", header[8]).group(1)) * u.km  # type: ignore
    )

    return VectorialModelResult(
        volume_density_grid=volume_density_grid,
        volume_density=volume_density,
        fragment_sputter=fragment_sputter,
        collision_sphere_radius=collision_sphere_radius,
        max_grid_radius=max_grid_radius,
        coma_radius=coma_radius,
    )


def write_rust_input_file(
    vmc: VectorialModelConfig, rust_input_filename: pathlib.Path
) -> None:
    """
    Takes a valid vectorial model config and produces a valid input file to give to the rust version of the model as input
    Only steady production is currently supported in the rust version
    """

    if vmc.production.time_variation is not None:
        print(
            "Only steady production is supported in the rust version! Running steady production model instead!."
        )

    # TODO: backflow mode radius, might require a re-definition of VectorialModelConfig
    rust_config_dict = {
        "base_q": float(vmc.production.base_q.to_value(1 / u.s)),  # type: ignore
        "p_tau_d": float(vmc.parent.tau_d.to_value(u.s)),  # type: ignore
        "p_tau_t": float(vmc.parent.tau_T.to_value(u.s)),  # type: ignore
        "v_outflow": float(vmc.parent.v_outflow.to_value(u.m / u.s)),  # type: ignore
        "sigma": float(vmc.parent.sigma.to_value(u.m**2)),  # type: ignore
        "f_tau_t": float(vmc.fragment.tau_T.to_value(u.s)),  # type: ignore
        "v_photo": float(vmc.fragment.v_photo.to_value(u.m / u.s)),  # type: ignore
        "radial_points": int(vmc.grid.radial_points),
        "angular_points": int(vmc.grid.angular_points),
        "radial_substeps": int(vmc.grid.radial_substeps),
        "parent_destruction_level": float(vmc.grid.parent_destruction_level),
        "fragment_destruction_level": float(vmc.grid.fragment_destruction_level),
    }

    with open(rust_input_filename, "w") as out_file:
        yaml.dump(rust_config_dict, out_file, default_flow_style=False)



import logging as log
import astropy.units as u

from scipy.interpolate import CubicSpline

from pyvectorial_au.model_output.vectorial_model_result import VectorialModelResult


def interpolate_volume_density(vmr: VectorialModelResult) -> None:
    if vmr.volume_density_interpolation is not None:
        log.debug(
            "Attempted to add volume density interpolation to a VectorialModelResult that already had one! Skipping."
        )
        return

    vmr.volume_density_interpolation = CubicSpline(
        vmr.volume_density_grid.to_value(u.m),  # type: ignore
        vmr.volume_density.to_value(1 / u.m**3),  # type: ignore
        bc_type="natural",
    )


def interpolate_column_density(vmr: VectorialModelResult) -> None:
    if vmr.column_density_interpolation is not None:
        log.debug(
            "Attempted to add column density interpolation to a VectorialModelResult that already had one! Skipping."
        )
        return

    if vmr.column_density_grid is None or vmr.column_density is None:
        log.debug(
            "Attempted to add column density interpolation to a VectorialModelResult that had no column density! Skipping."
        )
        return

    vmr.column_density_interpolation = CubicSpline(
        vmr.column_density_grid.to_value(u.m),  # type: ignore
        vmr.column_density.to_value(1 / u.m**2),  # type: ignore
        bc_type="natural",
    )



import logging as log
import numpy as np
import astropy.units as u

from abel.hansenlaw import hansenlaw_transform
from pyvectorial_au.model_output.vectorial_model_result import VectorialModelResult


def column_density_from_abel(vmr: VectorialModelResult, num_samples=10000) -> None:
    # we want to sample the volume density over the entire grid with a linspace:
    # the reverse Abel transform needs a constant delta between points
    sample_space_start = np.min(vmr.volume_density_grid)
    sample_space_end = np.max(vmr.volume_density_grid)
    sample_space, dr = np.linspace(
        start=sample_space_start.to_value(u.m),
        stop=sample_space_end.to_value(u.m),
        num=num_samples,
        retstep=True,
    )

    if vmr.volume_density_interpolation is None:
        log.debug(
            "Cannot add column density with Abel transform: no volume density interpolation available!"
        )
        return

    sampled_volume_densities = vmr.volume_density_interpolation(sample_space)

    abel_column_densities = hansenlaw_transform(
        sampled_volume_densities, dr=dr, direction="forward"
    )

    vmr.column_density_grid = sample_space * u.m
    vmr.column_density = abel_column_densities / u.m**2  # type: ignore



import numpy as np
import copy
import astropy.units as u
import scipy.interpolate
import plotly.graph_objects as go
from typing import Tuple
from typing import Optional

from pyvectorial_au.model_output.vectorial_model_result import VectorialModelResult
from pyvectorial_au.model_output.fragment_sputter import (
    FragmentSputterSpherical,
    FragmentSputterPolar,
)
from pyvectorial_au.post_model_processing.fragment_sputter_transform import (
    mirror_fragment_sputter,
    fragment_sputter_to_cartesian,
)


"""
    Functions for using plotly to plot various data contained in VectorialModelResults
"""

# color palette
myred = "#c74a77"
mybred = "#dbafad"
mygreen = "#afac7c"
mybgreen = "#dbd89c"
mypeach = "#dbb89c"
mybpeach = "#e9d4c3"
myblue = "#688894"
mybblue = "#a4b7be"
myblack = "#301e2a"
mybblack = "#82787f"
mywhite = "#d8d7dc"
mybwhite = "#e7e7ea"

# TODO: port over the column density inflection point marker in the matplotlib version?  Collision sphere marker
# function as well?

_default_vdens_units = 1 / u.m**3  # type: ignore


def plotly_volume_density_plot(
    vmr: VectorialModelResult,
    dist_units=u.m,
    vdens_units=_default_vdens_units,
    **kwargs
) -> go.Scatter:
    assert isinstance(vmr.volume_density, u.Quantity)
    assert isinstance(vmr.volume_density_grid, u.Quantity)

    xs = vmr.volume_density_grid.to(dist_units)
    ys = vmr.volume_density.to(vdens_units)

    vdens_plot = go.Scatter(x=xs, y=ys, **kwargs)
    return vdens_plot


def plotly_volume_density_interpolation_plot(
    vmr: VectorialModelResult,
    dist_units=u.m,
    vdens_units=_default_vdens_units,
    **kwargs
) -> go.Scatter:
    # model's interpolation function needs meters in, gives output in 1/m**3
    ys = (
        vmr.volume_density_interpolation(vmr.volume_density_grid.to_value(u.m)) / u.m**3  # type: ignore
    )

    vdens_interp_plot = go.Scatter(
        x=vmr.volume_density_grid.to_value(dist_units),  # type: ignore
        y=ys.to_value(vdens_units),
        **kwargs
    )
    return vdens_interp_plot


def plotly_column_density_plot(
    vmr: VectorialModelResult, dist_units=u.m, cdens_units=1 / u.m**2, **kwargs  # type: ignore
) -> go.Scatter:
    xs = vmr.column_density_grid.to(dist_units)  # type: ignore
    ys = vmr.column_density.to(cdens_units)  # type: ignore

    cdens_plot = go.Scatter(x=xs, y=ys, **kwargs)
    return cdens_plot


def plotly_column_density_interpolation_plot(
    vmr: VectorialModelResult, dist_units=u.m, cdens_units=1 / u.m**2, **kwargs  # type: ignore
) -> go.Scatter:
    # model's interpolation function needs meters in, gives output in 1/m**2
    ys = (
        vmr.column_density_interpolation(vmr.column_density_grid.to_value(u.m)) / u.m**2  # type: ignore
    )

    cdens_interp_plot = go.Scatter(
        x=vmr.column_density_grid.to_value(dist_units),  # type: ignore
        y=ys.to_value(cdens_units),
        **kwargs
    )
    return cdens_interp_plot


def plotly_column_density_plot_3d(
    vmr: VectorialModelResult,
    center=(0, 0) * u.km,  # type: ignore
    width=200000 * u.km,  # type: ignore
    height=200000 * u.km,  # type: ignore
    divisions=100,
    dist_units=u.m,
    cdens_units=1 / u.m**2,  # type: ignore
    **kwargs
) -> go.Surface:
    xmin_m, ymin_m = np.subtract(
        center.to_value(u.m), (width.to_value(u.m) / 2, height.to_value(u.m) / 2)
    )
    xmax_m, ymax_m = np.add(
        (xmin_m, ymin_m), (width.to_value(u.m), height.to_value(u.m))
    )
    xs_m = np.linspace(xmin_m, xmax_m, num=divisions)
    ys_m = np.linspace(ymin_m, ymax_m, num=divisions)

    x_mesh_m, y_mesh_m = np.meshgrid(xs_m, ys_m)
    z_mesh = (
        vmr.column_density_interpolation(np.sqrt(x_mesh_m**2 + y_mesh_m**2)) / u.m**2  # type: ignore
    )
    x_mesh = x_mesh_m * u.m
    y_mesh = y_mesh_m * u.m

    cdens_plot = go.Surface(
        x=x_mesh.to_value(dist_units),
        y=y_mesh.to_value(dist_units),
        z=z_mesh.to_value(cdens_units),
        **kwargs
    )
    return cdens_plot


def plotly_fragment_sputter_contour_plot(
    vmr,
    dist_units=u.km,
    sputter_units=1 / u.cm**3,
    within_r=2000 * u.km,  # type: ignore
    min_r=0 * u.km,  # type: ignore
    max_angle=np.pi,
    mirrored=False,
    show_outflow_axis=True,
    **kwargs
) -> Tuple[go.Contour, Optional[go.Scatter], float]:
    fragment_sputter = copy.deepcopy(vmr.fragment_sputter)

    if mirrored:
        fragment_sputter = mirror_fragment_sputter(fragment_sputter)

    within_max_angle = fragment_sputter.thetas < max_angle  # type: ignore
    fragment_sputter.rs = fragment_sputter.rs[within_max_angle]  # type: ignore
    fragment_sputter.thetas = fragment_sputter.thetas[within_max_angle]  # type: ignore
    fragment_sputter.fragment_density = fragment_sputter.fragment_density[
        within_max_angle
    ]

    if isinstance(fragment_sputter, FragmentSputterPolar) or isinstance(
        fragment_sputter, FragmentSputterSpherical
    ):
        fragment_sputter = fragment_sputter_to_cartesian(fragment_sputter)

    xs = fragment_sputter.xs
    ys = fragment_sputter.ys
    zs = fragment_sputter.fragment_density

    within_limit = np.logical_and(
        np.sqrt(xs**2 + ys**2) < within_r, np.sqrt(xs**2 + ys**2) > min_r
    )
    xs = xs[within_limit].to_value(dist_units)  # type: ignore
    ys = ys[within_limit].to_value(dist_units)  # type: ignore
    zs = zs[within_limit].to_value(sputter_units)  # type: ignore

    x_mesh, y_mesh = np.meshgrid(np.unique(xs), np.unique(ys))
    fs_rbf = scipy.interpolate.Rbf(xs, ys, zs, function="cubic")
    frag_mesh = fs_rbf(x_mesh, y_mesh)

    sputter_contour = go.Contour(
        x=np.unique(xs), y=np.unique(ys), z=frag_mesh, **kwargs
    )

    if show_outflow_axis:
        # highlight the outflow axis, along positive y
        outflow_axis = go.Scatter(
            x=[0, 0], y=[0, np.max(ys) * 0.9], mode="lines", opacity=0.5
        )
    else:
        outflow_axis = None

    return (sputter_contour, outflow_axis, np.max(xs))


def plotly_fragment_sputter_plot(
    vmr,
    dist_units=u.m,
    sputter_units=1 / u.m**3,  # type: ignore
    within_r=5000 * u.km,  # type: ignore
    min_r=0 * u.km,  # type: ignore
    mirrored=False,
    show_outflow_axis=True,
    **kwargs
) -> Tuple[Optional[go.Scatter3d], Optional[go.Scatter3d], Optional[float]]:
    fragment_sputter = copy.deepcopy(vmr.fragment_sputter)

    if mirrored:
        fragment_sputter = mirror_fragment_sputter(fragment_sputter)

    if isinstance(fragment_sputter, FragmentSputterPolar) or isinstance(
        fragment_sputter, FragmentSputterSpherical
    ):
        fragment_sputter = fragment_sputter_to_cartesian(fragment_sputter)

    # if mirrored:
    #     fragment_sputter = mirror_sputter(fragment_sputter)
    #
    # if isinstance(fragment_sputter, FragmentSputterPolar):
    #     fragment_sputter = cartesian_sputter_from_polar(fragment_sputter)

    xs = fragment_sputter.xs
    ys = fragment_sputter.ys
    zs = fragment_sputter.fragment_density

    # above_limit = np.sqrt(xs**2 + ys**2) > min_r
    # xs = xs[above_limit]
    # ys = ys[above_limit]
    # zs = zs[above_limit]

    within_limit = np.logical_and(
        np.sqrt(xs**2 + ys**2) < within_r, np.sqrt(xs**2 + ys**2) > min_r
    )
    if not len(xs):
        print("Radial cutoff for fragment sputter too small!  Nothing to plot.")
        return (None, None, None)

    xs = xs[within_limit].to_value(dist_units)  # type: ignore
    ys = ys[within_limit].to_value(dist_units)  # type: ignore
    zs = zs[within_limit].to_value(sputter_units)  # type: ignore

    sputter_plot = go.Scatter3d(
        x=xs, y=ys, z=zs, mode="markers", marker_color=zs, **kwargs
    )

    # highlight the outflow axis, along positive y
    if show_outflow_axis:
        outflow_axis = go.Scatter3d(
            x=[0, 0], y=[0, np.max(ys) * 0.9], z=[0, 0], mode="lines"
        )
    else:
        outflow_axis = None

    return (sputter_plot, outflow_axis, np.max(xs))


def plotly_q_t_plot(
    vmr: VectorialModelResult, time_units=u.hour
) -> Optional[go.Scatter]:
    if vmr.coma is None:
        return None

    # coma q_t function takes seconds, no astropy units attached
    ts_s = (np.linspace(-40, 40, num=1000) * u.hour).to_value(u.s)  # type: ignore
    f_q = np.vectorize(vmr.coma.q_t)
    qs = f_q(ts_s)
    t_h = (ts_s * u.s).to_value(time_units)

    qtplot = go.Scatter(x=t_h, y=qs, mode="lines")
    return qtplot



import pathlib
from typing import Optional

from pyvectorial_au.db.vectorial_model_cache import (
    get_vectorial_model_cache_session,
    VMCached,
)
from pyvectorial_au.model_input.vectorial_model_config import VectorialModelConfig
from pyvectorial_au.model_output.vectorial_model_calculation import EncodedVMCalculation
from pyvectorial_au.encoding.encoding_and_hashing import (
    vmc_to_sha256_digest,
    decompress_vmr_string,
    compress_vmr_string,
)


def get_saved_model_from_cache(
    vmc: VectorialModelConfig,
    vectorial_model_cache_path: pathlib.Path,
) -> Optional[EncodedVMCalculation]:
    """
    Given a vectorial model config, look up in the database at vectorial_model_cache_path
    whether that config has been run, and return the already-run model if so.
    """

    hashed_vmc = vmc_to_sha256_digest(vmc=vmc)
    # print(f"Looking up hashed vmc {hashed_vmc} in db ...")
    vm_cache_session = get_vectorial_model_cache_session(
        vectorial_model_cache_path=vectorial_model_cache_path
    )
    if vm_cache_session is None:
        print(f"No db session, skipping. [hash {hashed_vmc}]")
        return None

    cache_result = vm_cache_session.get(VMCached, hashed_vmc)
    if not cache_result:
        # print(f"Not found in db. [hash {hashed_vmc}]")
        vm_cache_session.close()
        return None

    evmc = EncodedVMCalculation(
        vmc=vmc,
        evmr=decompress_vmr_string(cache_result.vmr_b64_enc_zip),
        execution_time_s=cache_result.execution_time_s,
        vectorial_model_backend=cache_result.vectorial_model_backend,
        vectorial_model_version=cache_result.vectorial_model_version,
    )
    vm_cache_session.close()
    return evmc


def save_model_to_cache(
    evmc: EncodedVMCalculation, vectorial_model_cache_path: pathlib.Path
) -> None:
    """
    Given a completed model, check the given database at vectorial_model_cache_path if the given model
    already exists in the database.  If it does not, add it.
    """
    hashed_vmc = vmc_to_sha256_digest(vmc=evmc.vmc)
    # print(f"Saving model with hashed vmc {hashed_vmc} in db ...")

    vm_cache_session = get_vectorial_model_cache_session(
        vectorial_model_cache_path=vectorial_model_cache_path
    )
    if vm_cache_session is None:
        print(f"No db session, skipping. [hash {hashed_vmc}]")
        return

    with vm_cache_session.begin():
        model_exists = (
            vm_cache_session.query(VMCached).filter_by(vmc_hash=hashed_vmc).first()
        )
        if not model_exists:
            to_db = VMCached(
                vmc_hash=hashed_vmc,
                vmr_b64_enc_zip=compress_vmr_string(evmc.evmr),
                execution_time_s=evmc.execution_time_s,
                vectorial_model_backend=evmc.vectorial_model_backend,
                vectorial_model_version=evmc.vectorial_model_version,
            )
            vm_cache_session.add(to_db)
            vm_cache_session.commit()
